# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
===========================================
Array Module (:mod:`qiskit_dynamics.array`)
===========================================

.. currentmodule:: qiskit_dynamics.array

This module contains an :class:`Array` class that wraps N-dimensional array
objects from different libraries for use with Numpy functions through a common
interface and functions for working with Array objects and other libraries.

Array Class
===========

The :class:`Array` class provides a Numpy compatible wrapper to supported Python
array libraries. When applying Numpy functions to the array class these will be
dispatched to the corresponding function for the current Array backend, if a
compatible function has been specified in the :mod:`qiskit_dynamics.dispatch`
system.

The following array libraries have built in support for the
:mod:`qiskit_dynamics.dispatch` module

* `Numpy <https://numpy.org/>`_
* `JAX <https://github.com/google/jax>`_

Basic Usage
-----------

When using the default ``numpy`` backend :class:`Array` objects can be used
interchangable ``numpy.ndarray``. When ``numpy`` functions are applied to
an :class:`Array` object the return type will be an :class:`Array`
instead of an ``ndarray``

.. jupyter-execute::

    from qiskit_dynamics.array import Array
    import numpy as np

    # Initialize an Array
    a = Array(np.arange(10))

    # Apply numpy ufuncs
    np.cos(a) + 1j * np.sin(a)

For the JAX Array backend only Numpy functions that have a corresponding function
in the ``jax`` library that has been registed with the dispatch module can be
applied to the functions. Trying to apply an unsupported ``numpy`` function to
these arrays will raise an exception.


Default Backend
---------------

When initializing a new :class:`Array` the ``backend`` kwarg is used to specify
the array backend. This will convert the input data to an array of this backend
if required. If ``backend=None`` the default backend will be used.
The initial default backend is always set to ``"numpy"``.

The current default backend can be viewed by using the class method
:meth:`Array.default_backend`. A different
default backend can be set for all :meth:`Array` instances by using the
:meth:`Array.set_default_backend` class method.

Attributes and Methods
----------------------

The :class:`Array` class exposes the same methods and attributes of the wrapped
array class and adds two additional attributes

* :attr:`~Array.data` which returns the wrapped array object
* :attr:`~Array.backend` which returns the backend string of the wrapped array.

All other attributes and methods of the wrapped array are accessable
through this class, but with any array return types wrapped into
:class:`Array` objects. For example

.. jupyter-execute::

    # Call numpy.ndarray method
    a.reshape((2, 5))



Array Initialization
--------------------

An :class:`Array` object can be initialized as `Array(data)` from any ``data`` that is

1. An ``array`` object of one of the supported backends.
2. An object that can be used to initialize the backend array.
3. An object of a class that defines a ``__qiskit_array__`` method with the
   following signature

    .. code-block:: python

        def __qiskit_array__(self,
                             dtype: Optional[any] = None,
                             backend: Optional[str]=None) -> Array:
            # conversion from class to Array

The :meth:`Array.__init__` method has optional kwargs

* ``dtype`` (``any``): Equivalent to the ``numpy.array`` ``dtype`` kwarg.
  For other array backends the specified dtype must be supported on the
  array backend. If not specified the dytpe will be inferred from the
  input data.
* ``order`` (``str``): Equivalent to the ``numpy.array`` ``order`` kwarg.
  For other array backendsthe specified order value must be supported on
  the array backend.
* ``backend`` (``str``): The array backend to use. If not specified this
  will be inferred from the input data type if it is a backend array
  instance, otherwise it will be the :func:`default_backend`.


Using Arrays with other Libraries
=================================

The :class:`Array` class is intended to be used with array functions in
Python libraries that work with any of the supported Array backends.

Wrapping Functions
------------------

Functions from other libraries can be wrapped to work with Array objects using
the :func:`wrap` function as

.. code-block:: python

    from qiskit_dynamics.array import Array, wrap

    # Some 3rd-party library
    import library

    # Wrap library function f
    f = wrap(library.f)

    # Apply f to an Array
    a = Array([0, 1, 2, 4])
    b = f(a)

The wrapped function will automatically unwrapping any :class:`Array` args and kwargs
for calling the wrapped function. If the result of the function is a backend array
or an instance of a class that defines a ``__qiskit_array__`` method the wrapped
function will return this as an Array object. Similarly if the result is a tuple
this conversion will be applied to each element of the tuple where appropriate.

Wrapping Decorators
-------------------

The :func:`wrap` function can also be used to wrap function decorators by setting
the ``decorator=True`` kwarg.

For example to wrap autograd and jit decorators from the JAX library to work with
Array functions we can do the following

.. code-block:: python

    import jax

    jit = wrap(jax.jit, decorator=True)
    grad = wrap(jax.grad, decorator=True)
    value_and_grad = wrap(jax.value_and_grad, decorator=True)

    f = jit(obj)
    g = grad(obj)
    h = value_and_grad(obj)

.. note:

    Setting ``decorator=True`` when calling :func:`wrap` requires that the
    signature of the function being wrapped is
    ``func(f: Callable, ...) -> Callable``. Using it is equivalent to nested
    wrapping

    .. code-block:: python

        f_wrapped = wrap(func, decorator=True)(f)

    is equivalent to

    .. code-block:: python

        f_wrapped = wrap(wrap(func)(f))


Array Class
===========

.. autosummary::
    :toctree: ../stubs/

    Array

Array Functions
===============

.. autosummary::
    :toctree: ../stubs/

    wrap
"""


# Import Array
from .array import Array

# Import wrapper function
from .wrap import wrap

# Monkey patch quantum info
# pylint: disable= wrong-import-position
from .patch_qi import *



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Generator models module.
"""

from abc import ABC, abstractmethod
from typing import Tuple, Union, List, Optional
from warnings import warn
from copy import copy
import numpy as np
from scipy.sparse.csr import csr_matrix
from scipy.sparse import issparse, diags

from qiskit import QiskitError
from qiskit.quantum_info.operators import Operator
from qiskit_dynamics.models.operator_collections import (
    BaseOperatorCollection,
    DenseOperatorCollection,
    SparseOperatorCollection,
    JAXSparseOperatorCollection,
)
from qiskit_dynamics.array import Array
from qiskit_dynamics.signals import Signal, SignalList
from qiskit_dynamics.type_utils import to_numeric_matrix_type
from .rotating_frame import RotatingFrame

try:
    import jax
except ImportError:
    pass


class BaseGeneratorModel(ABC):
    r"""Defines an interface for a time-dependent
    linear differential equation of the form :math:`\dot{y}(t) = \Lambda(t, y)`,
    where :math:`\Lambda` is linear in :math:`y`. The core functionality is
    evaluation of :math:`\Lambda(t, y)`, as well as,
    if possible, evaluation of the map :math:`\Lambda(t, \cdot)`.

    Additionally, the class defines interfaces for:

     * Setting a "rotating frame", specified either directly as a :class:`RotatingFrame`
        instance, or an operator from which a :class:`RotatingFrame` instance can be constructed.
        The exact meaning of this transformation is determined by the structure of
        :math:`\Lambda(t, y)`, and is therefore by handled by concrete subclasses.
     * Setting an internal "evaluation mode", to set the specific numerical methods to use
        when evaluating :math:`\Lambda(t, y)`.
    """

    @property
    @abstractmethod
    def dim(self) -> int:
        """Gets matrix dimension."""

    @property
    @abstractmethod
    def rotating_frame(self) -> RotatingFrame:
        """Get the rotating frame."""

    @rotating_frame.setter
    @abstractmethod
    def rotating_frame(self, rotating_frame: RotatingFrame):
        """Set the rotating frame; either an already instantiated :class:`RotatingFrame` object,
        or a valid argument for the constructor of :class:`RotatingFrame`.
        """

    @property
    @abstractmethod
    def in_frame_basis(self) -> bool:
        """Whether or not the model is evaluated in the basis in which the frame is diagonalized."""

    @in_frame_basis.setter
    @abstractmethod
    def in_frame_basis(self, in_frame_basis: bool):
        """Set whether to evaluate in the basis in which the frame is diagonalized."""

    @property
    @abstractmethod
    def evaluation_mode(self) -> str:
        """Numerical evaluation mode of the model."""

    @evaluation_mode.setter
    @abstractmethod
    def evaluation_mode(self, new_mode: str):
        """Sets evaluation mode of model."""

    @abstractmethod
    def evaluate(self, time: float) -> Array:
        r"""If possible, evaluate the map :math:`\Lambda(t, \cdot)`.

        Args:
            time: Time.
        """

    @abstractmethod
    def evaluate_rhs(self, time: float, y: Array) -> Array:
        r"""Evaluate the right hand side :math:`\dot{y}(t) = \Lambda(t, y)`.

        Args:
            time: Time.
            y: State of the differential equation.
        """

    def copy(self):
        """Return a copy of self."""
        return copy(self)

    def __call__(self, time: float, y: Optional[Array] = None) -> Array:
        r"""Evaluate generator RHS functions. If ``y is None``,
        attemps to evaluate :math:`\Lambda(t, \cdot)`, otherwise, calculates
        :math:`\Lambda(t, y)`

        Args:
            time: Time.
            y: Optional state.

        Returns:
            Array: Either the evaluated model or the RHS for the given y
        """

        if y is None:
            return self.evaluate(time)

        return self.evaluate_rhs(time, y)


class GeneratorModel(BaseGeneratorModel):
    r"""A model for a a linear matrix differential equation in standard form.

    :class:`GeneratorModel` is a concrete instance of :class:`BaseGeneratorModel`, where the
    map :math:`\Lambda(t, y)` is explicitly constructed as:

    .. math::

        \Lambda(t, y) = G(t)y,

        G(t) = \sum_i s_i(t) G_i + G_d

    where the :math:`G_i` are matrices (represented by :class:`Operator`
    or :class:`Array` objects), the :math:`s_i(t)` are signals represented by
    a list of :class:`Signal` objects, and
    :math:`G_d` is the constant-in-time static term of the generator.
    """

    def __init__(
        self,
        static_operator: Optional[Array] = None,
        operators: Optional[Array] = None,
        signals: Optional[Union[SignalList, List[Signal]]] = None,
        rotating_frame: Optional[Union[Operator, Array, RotatingFrame]] = None,
        in_frame_basis: bool = False,
        evaluation_mode: str = "dense",
    ):
        """Initialize.

        Args:
            static_operator: Constant part of the generator.
            operators: A list of operators :math:`G_i`.
            signals: Stores the terms :math:`s_i(t)`. While required for evaluation,
                :class:`GeneratorModel` signals are not required at instantiation.
            rotating_frame: Rotating frame operator.
            in_frame_basis: Whether to represent the model in the basis in which the rotating
                frame operator is diagonalized.
            evaluation_mode: Evaluation mode to use.  Supported options are
                ``'dense'`` and ``'sparse'``. Call ``help(GeneratorModel.evaluation_mode)``
                for more details.
        Raises:
            QiskitError: If model not sufficiently specified.
        """
        if static_operator is None and operators is None:
            raise QiskitError(
                self.__class__.__name__
                + """ requires at least one of static_operator or operators to be
                              specified at construction."""
            )

        # initialize internal operator representation
        self._operator_collection = construct_operator_collection(
            evaluation_mode=evaluation_mode, static_operator=static_operator, operators=operators
        )
        self._evaluation_mode = evaluation_mode

        # set frame
        self._rotating_frame = None
        self.rotating_frame = rotating_frame

        # set operation in frame basis
        self._in_frame_basis = None
        self.in_frame_basis = in_frame_basis

        # initialize signal-related attributes
        self.signals = signals

    @property
    def dim(self) -> int:
        if self._operator_collection.static_operator is not None:
            return self._operator_collection.static_operator.shape[-1]
        else:
            return self._operator_collection.operators[0].shape[-1]

    @property
    def evaluation_mode(self) -> str:
        """Numerical evaluation mode of the model.

        Available options:

         * 'dense': Stores/evaluates operators using dense Arrays.
         * 'sparse': Stores/evaluates operators using sparse matrices. If
           the default Array backend is JAX, implemented with JAX BCOO arrays,
           otherwise uses scipy :class:`csr_matrix` sparse type. Note that
           JAX sparse mode is only recommended for use on CPU.
        """
        return self._evaluation_mode

    @evaluation_mode.setter
    def evaluation_mode(self, new_mode: str):
        """Set evaluation mode.

        Args:
            new_mode: String specifying new mode. Available options
                      are 'dense' and 'sparse'. See property doc string for details.

        Raises:
            NotImplementedError: if new_mode is not one of the above
            supported evaluation modes.
        """

        if new_mode != self._evaluation_mode:
            self._operator_collection = construct_operator_collection(
                new_mode,
                self._operator_collection.static_operator,
                self._operator_collection.operators,
            )
            self._evaluation_mode = new_mode

    @property
    def rotating_frame(self) -> RotatingFrame:
        """Return the rotating frame."""
        return self._rotating_frame

    @rotating_frame.setter
    def rotating_frame(self, rotating_frame: Union[Operator, Array, RotatingFrame]):
        new_frame = RotatingFrame(rotating_frame)

        new_static_operator = transfer_static_operator_between_frames(
            self._get_static_operator(in_frame_basis=True),
            new_frame=new_frame,
            old_frame=self.rotating_frame,
        )
        new_operators = transfer_operators_between_frames(
            self._get_operators(in_frame_basis=True),
            new_frame=new_frame,
            old_frame=self.rotating_frame,
        )

        self._rotating_frame = new_frame

        self._operator_collection = construct_operator_collection(
            self.evaluation_mode, new_static_operator, new_operators
        )

    @property
    def signals(self) -> SignalList:
        """Return the signals in the model."""
        return self._signals

    @signals.setter
    def signals(self, signals: Union[SignalList, List[Signal]]):
        """Set the signals."""

        if signals is None:
            self._signals = None
        elif signals is not None and self.operators is None:
            raise QiskitError("Signals must be None if operators is None.")
        else:
            # if signals is a list, instantiate a SignalList
            if isinstance(signals, list):
                signals = SignalList(signals)

            # if it isn't a SignalList by now, raise an error
            if not isinstance(signals, SignalList):
                raise QiskitError("Signals specified in unaccepted format.")

            # verify signal length is same as operators
            if isinstance(self.operators, list):
                len_operators = len(self.operators)
            else:
                len_operators = self.operators.shape[0]
            if len(signals) != len_operators:
                raise QiskitError("Signals needs to have the same length as operators.")

            self._signals = signals

    @property
    def in_frame_basis(self) -> bool:
        """Whether to represent the model in the basis in which the frame operator
        is diagonalized.
        """
        return self._in_frame_basis

    @in_frame_basis.setter
    def in_frame_basis(self, in_frame_basis: bool):
        self._in_frame_basis = in_frame_basis

    @property
    def operators(self) -> Array:
        """Get the operators in the model."""
        return self._get_operators(in_frame_basis=self._in_frame_basis)

    @property
    def static_operator(self) -> Array:
        """Get the static operator."""
        return self._get_static_operator(in_frame_basis=self._in_frame_basis)

    @static_operator.setter
    def static_operator(self, static_operator: Array):
        """Set the static operator."""
        self._set_static_operator(
            new_static_operator=static_operator, operator_in_frame_basis=self._in_frame_basis
        )

    def _get_operators(self, in_frame_basis: Optional[bool] = False) -> Array:
        """Get the operators used in the model construction.

        Args:
            in_frame_basis: Flag indicating whether to return the operators
            in the basis in which the rotating frame operator is diagonal.
        Returns:
            The operators in the basis specified by `in_frame_basis`.
        """
        ops = self._operator_collection.operators
        if ops is not None and not in_frame_basis and self.rotating_frame is not None:
            return self.rotating_frame.operator_out_of_frame_basis(ops)
        else:
            return ops

    def _get_static_operator(self, in_frame_basis: Optional[bool] = False) -> Array:
        """Get the constant term.

        Args:
            in_frame_basis: Flag for whether the returned static_operator should be
            in the basis in which the frame is diagonal.
        Returns:
            The static operator term.
        """
        op = self._operator_collection.static_operator
        if not in_frame_basis and self.rotating_frame is not None:
            return self.rotating_frame.operator_out_of_frame_basis(op)
        else:
            return op

    def _set_static_operator(
        self,
        new_static_operator: Array,
        operator_in_frame_basis: Optional[bool] = False,
    ):
        """Sets static term. Note that if the model has a rotating frame this will override
        any contributions to the static term due to the frame transformation.

        Args:
            new_static_operator: The static operator operator.
            operator_in_frame_basis: Whether `new_static_operator` is already in the rotating
            frame basis.
        """
        if new_static_operator is None:
            self._operator_collection.static_operator = None
        else:
            if not operator_in_frame_basis and self.rotating_frame is not None:
                new_static_operator = self.rotating_frame.operator_into_frame_basis(
                    new_static_operator
                )

            self._operator_collection.static_operator = new_static_operator

    def evaluate(self, time: float) -> Array:
        """Evaluate the model in array format as a matrix, independent of state.

        Args:
            time: Time to evaluate the model.

        Returns:
            Array: the evaluated model as a (n,n) matrix

        Raises:
            QiskitError: If model cannot be evaluated.
        """

        if self._signals is None:
            if self._operator_collection.operators is not None:
                raise QiskitError(
                    self.__class__.__name__
                    + " with non-empty operators cannot be evaluated without signals."
                )

            sig_vals = None
        else:
            sig_vals = self._signals.__call__(time)

        # Evaluated in frame basis, but without rotations
        op_combo = self._operator_collection(sig_vals)

        # Apply rotations e^{-Ft}Ae^{Ft} in frame basis where F = D
        return self.rotating_frame.operator_into_frame(
            time, op_combo, operator_in_frame_basis=True, return_in_frame_basis=self._in_frame_basis
        )

    def evaluate_rhs(self, time: float, y: Array) -> Array:
        r"""Evaluate `G(t) @ y`.

        Args:
            time: Time to evaluate the model.
            y: Array specifying system state.

        Returns:
            Array defined by :math:`G(t) \times y`.

        Raises:
            QiskitError: If model cannot be evaluated.
        """

        if self._signals is None:
            if self._operator_collection.operators is not None:
                raise QiskitError(
                    self.__class__.__name__
                    + " with non-empty operators cannot be evaluated without signals."
                )

            sig_vals = None
        else:
            sig_vals = self._signals.__call__(time)

        if self.rotating_frame is not None:
            # First, compute e^{tF}y as a pre-rotation in the frame basis
            out = self.rotating_frame.state_out_of_frame(
                time, y, y_in_frame_basis=self._in_frame_basis, return_in_frame_basis=True
            )
            # Then, compute the product Ae^{tF}y
            out = self._operator_collection(sig_vals, out)
            # Finally, we have the full operator e^{-tF}Ae^{tF}y
            out = self.rotating_frame.state_into_frame(
                time, out, y_in_frame_basis=True, return_in_frame_basis=self._in_frame_basis
            )
        else:
            return self._operator_collection(sig_vals, y)

        return out


def transfer_static_operator_between_frames(
    static_operator: Union[None, Array, csr_matrix],
    new_frame: Optional[Union[Array, RotatingFrame]] = None,
    old_frame: Optional[Union[Array, RotatingFrame]] = None,
) -> Tuple[Union[None, Array]]:
    """Helper function for transforming the static operator for a model from one
    frame basis into another.

    ``static_operator`` is additionally transformed as a generator: the old frame operator is
    added, and the new one is subtracted.

    Args:
        static_operator: Static operator of the model. None is treated as 0.
        new_frame: New rotating frame.
        old_frame: Old rotating frame.

    Returns:
        static_operator: Transformed as described above.
    """
    new_frame = RotatingFrame(new_frame)
    old_frame = RotatingFrame(old_frame)

    static_operator = to_numeric_matrix_type(static_operator)

    # transform out of old frame basis, and add the old frame operator
    if static_operator is not None:
        static_operator = old_frame.generator_out_of_frame(
            t=0.0,
            operator=static_operator,
            operator_in_frame_basis=True,
            return_in_frame_basis=False,
        )
    else:
        # "add" the frame operator to 0
        if old_frame.frame_operator is not None:
            if issparse(static_operator):
                if old_frame.frame_operator.ndim == 1:
                    static_operator = diags(old_frame.frame_operator, format="csr")
                else:
                    static_operator = csr_matrix(old_frame.frame_operator)
            else:
                if old_frame.frame_operator.ndim == 1:
                    static_operator = np.diag(old_frame.frame_operator)
                else:
                    static_operator = old_frame.frame_operator
    # transform into new frame basis, and add the new frame operator
    if static_operator is not None:
        static_operator = new_frame.generator_into_frame(
            t=0.0,
            operator=static_operator,
            operator_in_frame_basis=False,
            return_in_frame_basis=True,
        )
    else:
        # "subtract" the frame operator from 0
        if new_frame.frame_operator is not None:
            if issparse(static_operator):
                static_operator = -diags(new_frame.frame_diag, format="csr")
            else:
                static_operator = -np.diag(new_frame.frame_diag)

    return static_operator


def transfer_operators_between_frames(
    operators: Union[None, Array, List[csr_matrix]],
    new_frame: Optional[Union[Array, RotatingFrame]] = None,
    old_frame: Optional[Union[Array, RotatingFrame]] = None,
) -> Tuple[Union[None, Array]]:
    """Helper function for transforming a list of operators for a model
    from one frame basis into another.

    Args:
        operators: Operators of the model. If None, remains as None.
        new_frame: New rotating frame.
        old_frame: Old rotating frame.

    Returns:
        operators: Transformed as described above.
    """
    new_frame = RotatingFrame(new_frame)
    old_frame = RotatingFrame(old_frame)

    operators = to_numeric_matrix_type(operators)

    # transform out of old frame basis
    if operators is not None:
        # For sparse case, if list, loop
        if isinstance(operators, list):
            operators = [old_frame.operator_out_of_frame_basis(op) for op in operators]
        else:
            operators = old_frame.operator_out_of_frame_basis(operators)

    # transform into new frame basis
    if operators is not None:
        # For sparse case, if list, loop
        if isinstance(operators, list):
            operators = [new_frame.operator_into_frame_basis(op) for op in operators]
        else:
            operators = new_frame.operator_into_frame_basis(operators)

    return operators


def construct_operator_collection(
    evaluation_mode: str,
    static_operator: Union[None, Array, csr_matrix],
    operators: Union[None, Array, List[csr_matrix]],
) -> BaseOperatorCollection:
    """Construct operator collection for GeneratorModel.

    Args:
        evaluation_mode: Evaluation mode.
        static_operator: Static operator of the model.
        operators: Operators for the model.

    Returns:
        BaseOperatorCollection: The relevant operator collection.

    Raises:
        NotImplementedError: If the evaluation_mode is invalid.
    """

    if evaluation_mode == "dense":
        return DenseOperatorCollection(static_operator=static_operator, operators=operators)
    if evaluation_mode == "sparse" and Array.default_backend() == "jax":
        # warn that sparse mode when using JAX is primarily recommended for use on CPU
        if jax.default_backend() != "cpu":
            warn(
                """Using sparse mode with JAX is primarily recommended for use on CPU.""",
                stacklevel=2,
            )

        return JAXSparseOperatorCollection(static_operator=static_operator, operators=operators)
    if evaluation_mode == "sparse":
        return SparseOperatorCollection(static_operator=static_operator, operators=operators)

    raise NotImplementedError(
        "Evaluation mode '"
        + str(evaluation_mode)
        + "' is not supported. Call help(GeneratorModel.evaluation_mode) for available options."
    )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Operator collections as math/calculation objects for Model classes"""

from abc import ABC, abstractmethod
from typing import Union, List, Optional
from copy import copy
import numpy as np
from scipy.sparse import issparse
from scipy.sparse.csr import csr_matrix

from qiskit import QiskitError
from qiskit.quantum_info.operators.operator import Operator
from qiskit_dynamics.array import Array, wrap
from qiskit_dynamics.type_utils import to_array, to_csr, to_BCOO, vec_commutator, vec_dissipator

try:
    import jax.numpy as jnp
    from jax.experimental import sparse as jsparse

    # sparse versions of jax.numpy operations
    jsparse_sum = jsparse.sparsify(jnp.sum)
    jsparse_matmul = jsparse.sparsify(jnp.matmul)
    jsparse_add = jsparse.sparsify(jnp.add)
    jsparse_subtract = jsparse.sparsify(jnp.subtract)

    def jsparse_linear_combo(coeffs, mats):
        """Method for computing a linear combination of sparse arrays."""
        # pylint: disable=unexpected-keyword-arg
        return jsparse_sum(jnp.broadcast_to(coeffs[:, None, None], mats.shape) * mats, axis=0)

    # sparse version of computing A @ X @ B
    jsparse_triple_product = jsparse.sparsify(lambda A, X, B: A @ X @ B)

except ImportError:
    pass


class BaseOperatorCollection(ABC):
    r"""Abstract class representing a two-variable matrix function.

    This class represents a function :math:`c,y \mapsto \Lambda(c, y)`,
    which is assumed to be decomposed as
    :math:`\Lambda(c, y) = (G_d + \sum_jc_jG_j)  y`
    for matrices :math:`G_d` and :math:`G_j`, with
    :math:`G_d` referred to as the static operator.

    Describes an interface for evaluating the map or its action on ``y``,
    given the 1d set of values :math:`c_j`.
    """

    def __init__(
        self,
        static_operator: Optional[any] = None,
        operators: Optional[any] = None,
    ):
        """Initialize.

        Accepted types are determined by concrete subclasses.

        Args:
            operators: (k,n,n) Array specifying the terms :math:`G_j`.
            static_operator: (n,n) Array specifying the extra static_operator :math:`G_d`.
        """
        self.operators = operators
        self.static_operator = static_operator

    @property
    def static_operator(self) -> Array:
        """Returns static part of operator collection."""

    @static_operator.setter
    def static_operator(self, new_static_operator: Optional[Array] = None):
        """Sets static_operator term."""

    @property
    def operators(self) -> Array:
        """Return operators."""

    @operators.setter
    def operators(self, new_operators: Array) -> Array:
        """Return operators."""

    @abstractmethod
    def evaluate(self, signal_values: Array) -> Array:
        r"""Evaluate the map."""

    @abstractmethod
    def evaluate_rhs(self, signal_values: Union[List[Array], Array], y: Array) -> Array:
        r"""Compute the function."""

    def __call__(
        self, signal_values: Union[List[Array], Array], y: Optional[Array] = None
    ) -> Array:
        """Call either ``self.evaluate`` or ``self.evaluate_rhs`` depending on number of
        arguments.
        """

        if y is None:
            return self.evaluate(signal_values)

        return self.evaluate_rhs(signal_values, y)

    def copy(self):
        """Return a copy of self."""
        return copy(self)


class DenseOperatorCollection(BaseOperatorCollection):
    r"""Concrete operator collection representing a function computing left
    multiplication by an affine combination of matrices.

    Concrete instance of ``BaseOperatorCollection`` in which
    :math:`G_d` and :math:`G_j` are dense arrays.
    """

    @property
    def static_operator(self) -> Array:
        """Returns static part of operator collection."""
        return self._static_operator

    @static_operator.setter
    def static_operator(self, new_static_operator: Array):
        """Sets static_operator term."""
        self._static_operator = to_array(new_static_operator)

    @property
    def operators(self) -> Array:
        """Operators in the collection."""
        return self._operators

    @operators.setter
    def operators(self, new_operators: Array):
        self._operators = to_array(new_operators)

    def evaluate(self, signal_values: Union[Array, None]) -> Array:
        r"""Evaluate the affine combination of matrices.

        Returns:
            Evaluated model.
        Raises:
            QiskitError: if both static_operator and operators are None
        """
        if self._static_operator is not None and self._operators is not None:
            return np.tensordot(signal_values, self._operators, axes=1) + self._static_operator
        elif self._static_operator is None and self._operators is not None:
            return np.tensordot(signal_values, self._operators, axes=1)
        elif self._static_operator is not None:
            return self._static_operator
        else:
            raise QiskitError(
                self.__class__.__name__
                + """ with None for both static_operator and
                                operators cannot be evaluated."""
            )

    def evaluate_rhs(self, signal_values: Union[Array, None], y: Array) -> Array:
        """Evaluates the function."""
        return np.dot(self.evaluate(signal_values), y)


class SparseOperatorCollection(BaseOperatorCollection):
    r"""Sparse version of DenseOperatorCollection."""

    def __init__(
        self,
        static_operator: Optional[Union[Array, Operator]] = None,
        operators: Optional[Union[Array, List[Operator]]] = None,
        decimals: Optional[int] = 10,
    ):
        """Initialize.

        Args:
            static_operator: (n,n) Array specifying the static_operator term :math:`G_d`.
            operators: (k,n,n) Array specifying the terms :math:`G_j`.
            decimals: Values will be rounded at ``decimals`` places after decimal.
        """
        self._decimals = decimals
        super().__init__(static_operator=static_operator, operators=operators)

    @property
    def static_operator(self) -> csr_matrix:
        return self._static_operator

    @static_operator.setter
    def static_operator(self, new_static_operator: csr_matrix):
        if new_static_operator is not None:
            self._static_operator = np.round(to_csr(new_static_operator), self._decimals)
        else:
            self._static_operator = None

    @property
    def operators(self) -> List[csr_matrix]:
        if self._operators is None:
            return None

        return list(self._operators)

    @operators.setter
    def operators(self, new_operators: List[csr_matrix]):
        if new_operators is not None:
            new_operators_to_csr = to_csr(list(new_operators))
            new_operators = np.empty(shape=len(new_operators_to_csr), dtype="O")
            for idx, new_op in enumerate(new_operators_to_csr):
                new_operators[idx] = csr_matrix(np.round(new_op, self._decimals))

        self._operators = new_operators

    def evaluate(self, signal_values: Union[Array, None]) -> csr_matrix:
        r"""Sparse version of ``DenseOperatorCollection.evaluate``.

        Args:
            signal_values: Coefficients :math:`c_j`.

        Returns:
            Generator as sparse array.

        Raises:
            QiskitError: If collection cannot be evaluated.
        """
        if self._static_operator is not None and self._operators is not None:
            return (
                np.tensordot(signal_values, self._operators, axes=1).item() + self._static_operator
            )
        elif self._static_operator is None and self._operators is not None:
            return np.tensordot(signal_values, self._operators, axes=1).item()
        elif self.static_operator is not None:
            return self._static_operator
        raise QiskitError(
            self.__class__.__name__
            + """ with None for both static_operator and
                            operators cannot be evaluated."""
        )

    def evaluate_rhs(self, signal_values: Union[Array, None], y: Array) -> Array:
        if len(y.shape) == 2:
            # For 2d array, compute linear combination then multiply
            gen = self.evaluate(signal_values)
            return gen.dot(y)
        elif len(y.shape) == 1:
            # For a 1d array, multiply individual matrices then compute linear combination
            tmparr = np.empty(shape=(1), dtype="O")
            tmparr[0] = y

            if self._static_operator is not None and self._operators is not None:
                return np.dot(signal_values, self._operators * tmparr) + self.static_operator.dot(y)
            elif self._static_operator is None and self._operators is not None:
                return np.dot(signal_values, self._operators * tmparr)
            elif self.static_operator is not None:
                return self.static_operator.dot(y)

            raise QiskitError(
                self.__class__.__name__
                + """  with None for both static_operator and
                                operators cannot be evaluated."""
            )

        raise QiskitError(self.__class__.__name__ + """  cannot evaluate RHS for y.ndim > 3.""")


class JAXSparseOperatorCollection(BaseOperatorCollection):
    """Jax version of SparseOperatorCollection built on jax.experimental.sparse.BCOO."""

    @property
    def static_operator(self) -> "BCOO":
        return self._static_operator

    @static_operator.setter
    def static_operator(self, new_static_operator: Union["BCOO", None]):
        self._static_operator = to_BCOO(new_static_operator)

    @property
    def operators(self) -> Union["BCOO", None]:
        return self._operators

    @operators.setter
    def operators(self, new_operators: Union["BCOO", None]):
        self._operators = to_BCOO(new_operators)

    def evaluate(self, signal_values: Union[Array, None]) -> "BCOO":
        r"""Jax sparse version of ``DenseOperatorCollection.evaluate``.

        Args:
            signal_values: Coefficients :math:`c_j`.

        Returns:
            Generator as sparse jax array.

        Raises:
            QiskitError: If collection cannot be evaluated.
        """
        if signal_values is not None and isinstance(signal_values, Array):
            signal_values = signal_values.data

        if self._static_operator is not None and self._operators is not None:
            return jsparse_linear_combo(signal_values, self._operators) + self._static_operator
        elif self._static_operator is None and self._operators is not None:
            return jsparse_linear_combo(signal_values, self._operators)
        elif self.static_operator is not None:
            return self._static_operator
        raise QiskitError(
            self.__class__.__name__
            + """ with None for both static_operator and
                            operators cannot be evaluated."""
        )

    def evaluate_rhs(self, signal_values: Union[Array, None], y: Array) -> Array:
        if y.ndim < 3:
            if isinstance(y, Array):
                y = y.data
            return Array(jsparse_matmul(self.evaluate(signal_values), y))

        raise QiskitError(self.__class__.__name__ + """  cannot evaluate RHS for y.ndim >= 3.""")


class BaseLindbladOperatorCollection(ABC):
    r"""Abstract class representing a two-variable matrix function for evaluating
    the right hand side of the Lindblad equation.

    In particular, this object represents the function:
        .. math::
            \Lambda(c_1, c_2, \rho) = -i[H_d + \sum_j c_{1,j}H_j,\rho]
                                        + \sum_j(D_j\rho D_j^\dagger
                                          - (1/2) * {D_j^\daggerD_j,\rho})
                                      + \sum_jc_{2,j}(L_j\rho L_j^\dagger
                                        - (1/2) * {L_j^\daggerL_j,\rho})

    where :math:`\[,\]` and :math:`\{,\}` are the operator
    commutator and anticommutator, respectively.

    Describes an interface for evaluating the map or its action on :math:`\rho`,
    given a pair of 1d sets of values :math:`c_1, c_2`.
    """

    def __init__(
        self,
        static_hamiltonian: Optional[any] = None,
        hamiltonian_operators: Optional[any] = None,
        static_dissipators: Optional[any] = None,
        dissipator_operators: Optional[any] = None,
    ):
        r"""Initialize collection. Argument types depend on concrete subclass.

        Args:
            static_hamiltonian: Constant term :math:`H_d` to be added to the Hamiltonian of the
                                system.
            hamiltonian_operators: Specifies breakdown of Hamiltonian
                as :math:`H(t) = \sum_j s(t) H_j+H_d` by specifying H_j. (k,n,n) array.
            static_dissipators: Constant dissipator terms.
            dissipator_operators: the terms :math:`L_j` in Lindblad equation. (m,n,n) array.
        """
        self.static_hamiltonian = static_hamiltonian
        self.hamiltonian_operators = hamiltonian_operators
        self.static_dissipators = static_dissipators
        self.dissipator_operators = dissipator_operators

    @property
    @abstractmethod
    def static_hamiltonian(self) -> Array:
        """Returns static part of the hamiltonian."""

    @static_hamiltonian.setter
    @abstractmethod
    def static_hamiltonian(self, new_static_operator: Optional[Array] = None):
        """Sets static_operator term."""

    @property
    @abstractmethod
    def hamiltonian_operators(self) -> Array:
        """Returns operators for non-static part of Hamiltonian."""

    @hamiltonian_operators.setter
    @abstractmethod
    def hamiltonian_operators(self, new_hamiltonian_operators: Optional[Array] = None):
        """Set operators for non-static part of Hamiltonian."""

    @property
    @abstractmethod
    def static_dissipators(self) -> Array:
        """Returns operators for static part of dissipator."""

    @static_dissipators.setter
    @abstractmethod
    def static_dissipators(self, new_static_dissipators: Optional[Array] = None):
        """Sets operators for static part of dissipator."""

    @property
    @abstractmethod
    def dissipator_operators(self) -> Array:
        """Returns operators for non-static part of dissipator."""

    @dissipator_operators.setter
    @abstractmethod
    def dissipator_operators(self, new_dissipator_operators: Optional[Array] = None):
        """Sets operators for non-static part of dissipator."""

    @abstractmethod
    def evaluate_hamiltonian(self, ham_sig_vals: Union[None, Array]) -> Union[csr_matrix, Array]:
        """Evaluate the Hamiltonian of the model."""

    @abstractmethod
    def evaluate(
        self, ham_sig_vals: Union[None, Array], dis_sig_vals: Union[None, Array]
    ) -> Union[csr_matrix, Array]:
        r"""Evaluate the map."""

    @abstractmethod
    def evaluate_rhs(
        self, ham_sig_vals: Union[None, Array], dis_sig_vals: Union[None, Array], y: Array
    ) -> Array:
        r"""Compute the function."""

    def __call__(
        self, ham_sig_vals: Union[None, Array], dis_sig_vals: Union[None, Array], y: Optional[Array]
    ) -> Union[csr_matrix, Array]:
        """Evaluate the model, or evaluate the RHS."""
        if y is None:
            return self.evaluate(ham_sig_vals, dis_sig_vals)

        return self.evaluate_rhs(ham_sig_vals, dis_sig_vals, y)

    def copy(self):
        """Return a copy of self."""
        return copy(self)


class DenseLindbladCollection(BaseLindbladOperatorCollection):
    r"""Object for computing the right hand side of the Lindblad equation
    with dense arrays.
    """

    @property
    def static_hamiltonian(self) -> Array:
        return self._static_hamiltonian

    @static_hamiltonian.setter
    def static_hamiltonian(self, new_static_hamiltonian: Optional[Array] = None):
        self._static_hamiltonian = to_array(new_static_hamiltonian)

    @property
    def hamiltonian_operators(self) -> Array:
        return self._hamiltonian_operators

    @hamiltonian_operators.setter
    def hamiltonian_operators(self, new_hamiltonian_operators: Optional[Array] = None):
        self._hamiltonian_operators = to_array(new_hamiltonian_operators)

    @property
    def static_dissipators(self) -> Array:
        return self._static_dissipators

    @static_dissipators.setter
    def static_dissipators(self, new_static_dissipators: Optional[Array] = None):
        self._static_dissipators = to_array(new_static_dissipators)
        if self._static_dissipators is not None:
            self._static_dissipators_adj = np.conjugate(
                np.transpose(self._static_dissipators, [0, 2, 1])
            ).copy()
            self._static_dissipators_product_sum = -0.5 * np.sum(
                np.matmul(self._static_dissipators_adj, self._static_dissipators), axis=0
            )

    @property
    def dissipator_operators(self) -> Array:
        return self._dissipator_operators

    @dissipator_operators.setter
    def dissipator_operators(self, new_dissipator_operators: Optional[Array] = None):
        self._dissipator_operators = to_array(new_dissipator_operators)
        if self._dissipator_operators is not None:
            self._dissipator_operators_adj = np.conjugate(
                np.transpose(self._dissipator_operators, [0, 2, 1])
            ).copy()
            self._dissipator_products = np.matmul(
                self._dissipator_operators_adj, self._dissipator_operators
            )

    def evaluate(self, ham_sig_vals: Array, dis_sig_vals: Array) -> Array:
        raise ValueError("Non-vectorized Lindblad collections cannot be evaluated without a state.")

    def evaluate_hamiltonian(self, ham_sig_vals: Union[None, Array]) -> Array:
        r"""Compute the Hamiltonian.

        Args:
            ham_sig_vals: [Real] values of :math:`s_j` in :math:`H = \sum_j s_j(t) H_j + H_d`.
        Returns:
            Hamiltonian matrix.
        Raises:
            QiskitError: If collection not sufficiently specified.
        """
        if self._static_hamiltonian is not None and self._hamiltonian_operators is not None:
            return (
                np.tensordot(ham_sig_vals, self._hamiltonian_operators, axes=1)
                + self._static_hamiltonian
            )
        elif self._static_hamiltonian is None and self._hamiltonian_operators is not None:
            return np.tensordot(ham_sig_vals, self._hamiltonian_operators, axes=1)
        elif self._static_hamiltonian is not None:
            return self._static_hamiltonian
        else:
            raise QiskitError(
                self.__class__.__name__
                + """ with None for both static_hamiltonian and
                                hamiltonian_operators cannot evaluate Hamiltonian."""
            )

    def evaluate_rhs(
        self, ham_sig_vals: Union[None, Array], dis_sig_vals: Union[None, Array], y: Array
    ) -> Array:
        r"""Evaluates Lindblad equation RHS given a pair of signal values
        for the hamiltonian terms and the dissipator terms. Expresses
        the RHS of the Lindblad equation as :math:`(A+B)y + y(A-B) + C`, where
            .. math::
            A = (-1/2)*\sum_jD_j^\dagger D_j + (-1/2)*\sum_j\gamma_j(t) L_j^\dagger L_j,

            B = -iH,

            C = \sum_j \gamma_j(t) L_j y L_j^\dagger.

        Args:
            ham_sig_vals: hamiltonian coefficient values, :math:`s_j(t)`.
            dis_sig_vals: dissipator signal values, :math:`\gamma_j(t)`.
            y: density matrix as (n,n) Array representing the state at time :math:`t`.
        Returns:
            RHS of Lindblad equation
            .. math::
                -i[H,y] + \sum_j\gamma_j(t)(L_j y L_j^\dagger - (1/2) * {L_j^\daggerL_j,y}).
        Raises:
            QiskitError: If operator collection is underspecified.
        """

        hamiltonian_matrix = None
        if self._static_hamiltonian is not None or self._hamiltonian_operators is not None:
            hamiltonian_matrix = -1j * self.evaluate_hamiltonian(ham_sig_vals)  # B matrix

        # if dissipators present (includes both hamiltonian is None and is not None)
        if self._dissipator_operators is not None or self._static_dissipators is not None:

            # A matrix
            if self._static_dissipators is None:
                dissipators_matrix = np.tensordot(
                    -0.5 * dis_sig_vals, self._dissipator_products, axes=1
                )
            elif self._dissipator_operators is None:
                dissipators_matrix = self._static_dissipators_product_sum
            else:
                dissipators_matrix = self._static_dissipators_product_sum + np.tensordot(
                    -0.5 * dis_sig_vals, self._dissipator_products, axes=1
                )

            if hamiltonian_matrix is not None:
                left_mult_contribution = np.matmul(hamiltonian_matrix + dissipators_matrix, y)
                right_mult_contribution = np.matmul(y, dissipators_matrix - hamiltonian_matrix)
            else:
                left_mult_contribution = np.matmul(dissipators_matrix, y)
                right_mult_contribution = np.matmul(y, dissipators_matrix)

            if len(y.shape) == 3:
                # Must do array broadcasting and transposition to ensure vectorization works
                y = np.broadcast_to(y, (1, y.shape[0], y.shape[1], y.shape[2])).transpose(
                    [1, 0, 2, 3]
                )

            if self._static_dissipators is None:
                both_mult_contribution = np.tensordot(
                    dis_sig_vals,
                    np.matmul(
                        self._dissipator_operators, np.matmul(y, self._dissipator_operators_adj)
                    ),
                    axes=(-1, -3),
                )
            elif self._dissipator_operators is None:
                both_mult_contribution = np.sum(
                    np.matmul(self._static_dissipators, np.matmul(y, self._static_dissipators_adj)),
                    axis=-3,
                )
            else:
                both_mult_contribution = np.sum(
                    np.matmul(self._static_dissipators, np.matmul(y, self._static_dissipators_adj)),
                    axis=-3,
                ) + np.tensordot(
                    dis_sig_vals,
                    np.matmul(
                        self._dissipator_operators, np.matmul(y, self._dissipator_operators_adj)
                    ),
                    axes=(-1, -3),
                )

            return left_mult_contribution + right_mult_contribution + both_mult_contribution
        # if just hamiltonian
        elif hamiltonian_matrix is not None:
            return np.dot(hamiltonian_matrix, y) - np.dot(y, hamiltonian_matrix)
        else:
            raise QiskitError(
                """DenseLindbladCollection with None for static_hamiltonian,
                                 hamiltonian_operators, static_dissipators, and
                                 dissipator_operators, cannot evaluate rhs."""
            )


class SparseLindbladCollection(DenseLindbladCollection):
    """Sparse version of DenseLindbladCollection."""

    def __init__(
        self,
        static_hamiltonian: Optional[Union[csr_matrix, Operator]] = None,
        hamiltonian_operators: Optional[Union[List[csr_matrix], List[Operator]]] = None,
        static_dissipators: Optional[Union[List[csr_matrix], List[Operator]]] = None,
        dissipator_operators: Optional[Union[List[csr_matrix], List[Operator]]] = None,
        decimals: Optional[int] = 10,
    ):
        r"""Initializes sparse lindblad collection.

        Args:
            static_hamiltonian: Constant term :math:`H_d` to be added to the Hamiltonian of the
                                system.
            hamiltonian_operators: Specifies breakdown of Hamiltonian
                as :math:`H(t) = \sum_j s(t) H_j+H_d` by specifying H_j. (k,n,n) array.
            dissipator_operators: the terms :math:`L_j` in Lindblad equation. (m,n,n) array.
            decimals: operator values will be rounded to ``decimals`` places after the
                decimal place to avoid excess storage of near-zero values
                in sparse format.
        """
        self._decimals = decimals
        super().__init__(
            static_hamiltonian=static_hamiltonian,
            hamiltonian_operators=hamiltonian_operators,
            static_dissipators=static_dissipators,
            dissipator_operators=dissipator_operators,
        )

    @property
    def static_hamiltonian(self) -> csr_matrix:
        return self._static_hamiltonian

    @static_hamiltonian.setter
    def static_hamiltonian(self, new_static_hamiltonian: Optional[csr_matrix] = None):
        if new_static_hamiltonian is not None:
            new_static_hamiltonian = np.round(
                to_csr(new_static_hamiltonian), decimals=self._decimals
            )
        self._static_hamiltonian = new_static_hamiltonian

    @property
    def hamiltonian_operators(self) -> np.ndarray:
        if self._hamiltonian_operators is None:
            return None

        return list(self._hamiltonian_operators)

    @hamiltonian_operators.setter
    def hamiltonian_operators(self, new_hamiltonian_operators: Optional[List[csr_matrix]] = None):
        if new_hamiltonian_operators is not None:
            new_hamiltonian_operators = to_csr(new_hamiltonian_operators)
            new_hamiltonian_operators = [
                np.round(op, decimals=self._decimals) for op in new_hamiltonian_operators
            ]
            new_hamiltonian_operators = np.array(new_hamiltonian_operators, dtype="O")

        self._hamiltonian_operators = new_hamiltonian_operators

    @property
    def static_dissipators(self) -> Union[None, csr_matrix]:
        if self._static_dissipators is None:
            return None

        return list(self._static_dissipators)

    @static_dissipators.setter
    def static_dissipators(self, new_static_dissipators: Optional[List[csr_matrix]] = None):
        """Set up the dissipators themselves, as well as their adjoints, and the product of
        adjoint with operator.
        """
        self._static_dissipators = None
        if new_static_dissipators is not None:
            # setup new dissipators
            new_static_dissipators = to_csr(new_static_dissipators)
            new_static_dissipators = [
                np.round(op, decimals=self._decimals) for op in new_static_dissipators
            ]

            # setup adjoints
            static_dissipators_adj = [op.conj().transpose() for op in new_static_dissipators]

            # wrap in object arrays
            new_static_dissipators = np.array(new_static_dissipators, dtype="O")
            static_dissipators_adj = np.array(static_dissipators_adj, dtype="O")

            # pre-compute products
            static_dissipators_product_sum = -0.5 * np.sum(
                static_dissipators_adj * new_static_dissipators, axis=0
            )

            self._static_dissipators = new_static_dissipators
            self._static_dissipators_adj = static_dissipators_adj
            self._static_dissipators_product_sum = static_dissipators_product_sum

    @property
    def dissipator_operators(self) -> Union[None, List[csr_matrix]]:
        if self._dissipator_operators is None:
            return None

        return list(self._dissipator_operators)

    @dissipator_operators.setter
    def dissipator_operators(self, new_dissipator_operators: Optional[List[csr_matrix]] = None):
        """Set up the dissipators themselves, as well as their adjoints, and the product of
        adjoint with operator.
        """
        self._dissipator_operators = None
        if new_dissipator_operators is not None:
            # setup new dissipators
            new_dissipator_operators = to_csr(new_dissipator_operators)
            new_dissipator_operators = [
                np.round(op, decimals=self._decimals) for op in new_dissipator_operators
            ]

            # setup adjoints
            dissipator_operators_adj = [op.conj().transpose() for op in new_dissipator_operators]

            # wrap in object arrays
            new_dissipator_operators = np.array(new_dissipator_operators, dtype="O")
            dissipator_operators_adj = np.array(dissipator_operators_adj, dtype="O")

            # pre-compute projducts
            dissipator_products = dissipator_operators_adj * new_dissipator_operators

            self._dissipator_operators = new_dissipator_operators
            self._dissipator_operators_adj = dissipator_operators_adj
            self._dissipator_products = dissipator_products

    def evaluate_hamiltonian(self, ham_sig_vals: Union[None, Array]) -> csr_matrix:
        r"""Compute the Hamiltonian.

        Args:
            ham_sig_vals: [Real] values of :math:`s_j` in :math:`H = \sum_j s_j(t) H_j + H_d`.
        Returns:
            Hamiltonian matrix.
        Raises:
            QiskitError: If collection not sufficiently specified.
        """
        if self._static_hamiltonian is not None and self._hamiltonian_operators is not None:
            return (
                np.sum(ham_sig_vals * self._hamiltonian_operators, axis=-1)
                + self.static_hamiltonian
            )
        elif self._static_hamiltonian is None and self._hamiltonian_operators is not None:
            return np.sum(ham_sig_vals * self._hamiltonian_operators, axis=-1)
        elif self._static_hamiltonian is not None:
            return self._static_hamiltonian
        else:
            raise QiskitError(
                self.__class__.__name__
                + """ with None for both static_hamiltonian and
                                hamiltonian_operators cannot evaluate Hamiltonian."""
            )

    def evaluate_rhs(
        self, ham_sig_vals: Union[None, Array], dis_sig_vals: Union[None, Array], y: Array
    ) -> Array:
        r"""Evaluates the RHS of the LindbladModel for a given list of signal values.

        Args:
            ham_sig_vals: stores Hamiltonian signal values :math:`s_j(t)`.
            dis_sig_vals: stores dissipator signal values :math:`\gamma_j(t)`.
                Pass None if no dissipator operators involved.
            y: density matrix of system. (k,n,n) Array.
        Returns:
            RHS of Lindbladian.
        Raises:
            QiskitError: If RHS cannot be evaluated due to insufficient collection data.

        Calculation details:
            * for csr_matrices is equivalent to matrix multiplicaiton.
            We use numpy array broadcasting rules, combined with the above
            fact, to achieve speeds that are substantially faster than a for loop.
            First, in the case of a single (n,n) density matrix, we package the entire
            array as a single-element array whose entry is the array. In the case of
            multiple density matrices a (k,n,n) Array, we package everything as a
            (k,1) Array whose [j,0] entry is the [j,:,:] density matrix.

            In calculating the left- and right-mult contributions, we package
            H+L and H-L as (1) object arrays whose single entry stores the relevant
            sparse matrix. We can then multiply our packaged density matrix and
            [H\pm L]. Using numpy broadcasting rules, [H\pm L] will be broadcast
            to a (k,1) Array for elementwise multiplication with our packaged density
            matrices. After this, elementwise multiplication is applied. This in turn
            references each object's __mul__ function, which–for our csr_matrix components
            means matrix multiplication.

            In calculating the left-right-multiplication part, we use our (m)-shape
            object arrays holding the dissipator operators to perform multiplication.
            We can take an elementwise product with our packaged density matrix, at which
            point our dissipator operators are broadcast as (m) -> (1,m) -> (k,m) shaped,
            and our packaged density matrix as (k,1) -> (k,m). Elementwise multiplication
            is then applied, which is interpreted as matrix multiplication. This yields
            an array where entry [i,j] is an object storing the results of s_jL_j\rho_i L_j^\dagger.
            We can then sum over j and unpackage our object array to get our desired result.
        """
        hamiltonian_matrix = None
        if self._static_hamiltonian is not None or self._hamiltonian_operators is not None:
            hamiltonian_matrix = -1j * self.evaluate_hamiltonian(ham_sig_vals)  # B matrix

        # package (n,n) Arrays as (1)
        # Arrays of dtype object, or (k,n,n) Arrays as (k,1) Arrays of dtype object
        y = package_density_matrices(y)

        # if dissipators present (includes both hamiltonian is None and is not None)
        if self._dissipator_operators is not None or self._static_dissipators is not None:

            # A matrix
            if self._static_dissipators is None:
                dissipators_matrix = np.sum(
                    -0.5 * dis_sig_vals * self._dissipator_products, axis=-1
                )
            elif self._dissipator_operators is None:
                dissipators_matrix = self._static_dissipators_product_sum
            else:
                dissipators_matrix = self._static_dissipators_product_sum + np.sum(
                    -0.5 * dis_sig_vals * self._dissipator_products, axis=-1
                )

            if hamiltonian_matrix is not None:
                left_mult_contribution = np.squeeze([hamiltonian_matrix + dissipators_matrix] * y)
                right_mult_contribution = np.squeeze(y * [dissipators_matrix - hamiltonian_matrix])
            else:
                left_mult_contribution = np.squeeze([dissipators_matrix] * y)
                right_mult_contribution = np.squeeze(y * [dissipators_matrix])

            # both_mult_contribution[i] = \gamma_i L_i\rho L_i^\dagger performed in array language
            if self._static_dissipators is None:
                both_mult_contribution = np.sum(
                    (dis_sig_vals * self._dissipator_operators)
                    * y
                    * self._dissipator_operators_adj,
                    axis=-1,
                )
            elif self._dissipator_operators is None:
                both_mult_contribution = np.sum(
                    self._static_dissipators * y * self._static_dissipators_adj, axis=-1
                )
            else:
                both_mult_contribution = np.sum(
                    (dis_sig_vals * self._dissipator_operators)
                    * y
                    * self._dissipator_operators_adj,
                    axis=-1,
                ) + np.sum(self._static_dissipators * y * self._static_dissipators_adj, axis=-1)

            out = left_mult_contribution + right_mult_contribution + both_mult_contribution

        elif hamiltonian_matrix is not None:
            out = (([hamiltonian_matrix] * y) - (y * [hamiltonian_matrix]))[0]
        else:
            raise QiskitError(
                "SparseLindbladCollection with None for static_hamiltonian, "
                "hamiltonian_operators, and dissipator_operators, cannot evaluate rhs."
            )
        if len(y.shape) == 2:
            # Very slow; avoid if not necessary (or if better implementation found). Needs to
            # map a (k) Array of dtype object with j^{th} entry a (n,n) Array -> (k,n,n) Array.
            out = unpackage_density_matrices(out.reshape(y.shape[0], 1))

        return out


class JAXSparseLindbladCollection(BaseLindbladOperatorCollection):
    r"""Object for computing the right hand side of the Lindblad equation
    with using jax.experimental.sparse.BCOO arrays.
    """

    @property
    def static_hamiltonian(self) -> "BCOO":
        return self._static_hamiltonian

    @static_hamiltonian.setter
    def static_hamiltonian(self, new_static_hamiltonian: Union["BCOO", None]):
        self._static_hamiltonian = to_BCOO(new_static_hamiltonian)

    @property
    def hamiltonian_operators(self) -> Union["BCOO", None]:
        return self._hamiltonian_operators

    @hamiltonian_operators.setter
    def hamiltonian_operators(self, new_hamiltonian_operators: Union["BCOO", None]):
        self._hamiltonian_operators = to_BCOO(new_hamiltonian_operators)

    @property
    def static_dissipators(self) -> Union["BCOO", None]:
        return self._static_dissipators

    @static_dissipators.setter
    def static_dissipators(self, new_static_dissipators: Union["BCOO", None]):
        """Operators constructed using dense operations."""
        self._static_dissipators = to_array(new_static_dissipators)
        if self._static_dissipators is not None:
            self._static_dissipators_adj = np.conjugate(
                np.transpose(self._static_dissipators, [0, 2, 1])
            ).copy()
            self._static_dissipators_product_sum = -0.5 * np.sum(
                np.matmul(self._static_dissipators_adj, self._static_dissipators), axis=0
            )
            self._static_dissipators = jsparse.BCOO.fromdense(
                self._static_dissipators.data, n_batch=1
            )
            self._static_dissipators_adj = jsparse.BCOO.fromdense(
                self._static_dissipators_adj.data, n_batch=1
            )
            self._static_dissipators_product_sum = jsparse.BCOO.fromdense(
                self._static_dissipators_product_sum.data
            )

    @property
    def dissipator_operators(self) -> Union["BCOO", None]:
        return self._dissipator_operators

    @dissipator_operators.setter
    def dissipator_operators(self, new_dissipator_operators: Union["BCOO", None]):
        """Operators constructed using dense operations."""
        self._dissipator_operators = to_array(new_dissipator_operators)
        if self._dissipator_operators is not None:
            self._dissipator_operators_adj = np.conjugate(
                np.transpose(self._dissipator_operators, [0, 2, 1])
            ).copy()
            self._dissipator_products = np.matmul(
                self._dissipator_operators_adj, self._dissipator_operators
            )
            self._dissipator_operators = jsparse.BCOO.fromdense(
                self._dissipator_operators.data, n_batch=1
            )
            self._dissipator_operators_adj = jsparse.BCOO.fromdense(
                self._dissipator_operators_adj.data, n_batch=1
            )
            self._dissipator_products = -0.5 * jsparse.BCOO.fromdense(
                self._dissipator_products.data, n_batch=1
            )

    def evaluate(self, ham_sig_vals: Array, dis_sig_vals: Array) -> Array:
        raise ValueError("Non-vectorized Lindblad collections cannot be evaluated without a state.")

    def evaluate_hamiltonian(self, ham_sig_vals: Union["BCOO", None]) -> "BCOO":
        r"""Compute the Hamiltonian.

        Args:
            ham_sig_vals: [Real] values of :math:`s_j` in :math:`H = \sum_j s_j(t) H_j + H_d`.
        Returns:
            Hamiltonian matrix.
        Raises:
            QiskitError: If collection not sufficiently specified.
        """
        if isinstance(ham_sig_vals, Array):
            ham_sig_vals = ham_sig_vals.data

        if self._static_hamiltonian is not None and self._hamiltonian_operators is not None:
            return (
                jsparse_linear_combo(ham_sig_vals, self._hamiltonian_operators)
                + self._static_hamiltonian
            )
        elif self._static_hamiltonian is None and self._hamiltonian_operators is not None:
            return jsparse_linear_combo(ham_sig_vals, self._hamiltonian_operators)
        elif self._static_hamiltonian is not None:
            return self._static_hamiltonian
        else:
            raise QiskitError(
                self.__class__.__name__
                + """ with None for both static_hamiltonian and
                                hamiltonian_operators cannot evaluate Hamiltonian."""
            )

    @wrap
    def evaluate_rhs(
        self, ham_sig_vals: Union[None, Array], dis_sig_vals: Union[None, Array], y: Array
    ) -> Array:
        r"""Evaluates Lindblad equation RHS given a pair of signal values
        for the hamiltonian terms and the dissipator terms. Expresses
        the RHS of the Lindblad equation as :math:`(A+B)y + y(A-B) + C`, where
            .. math::
            A = (-1/2)*\sum_jD_j^\dagger D_j + (-1/2)*\sum_j\gamma_j(t) L_j^\dagger L_j,

            B = -iH,

            C = \sum_j \gamma_j(t) L_j y L_j^\dagger.

        Args:
            ham_sig_vals: hamiltonian coefficient values, :math:`s_j(t)`.
            dis_sig_vals: dissipator signal values, :math:`\gamma_j(t)`.
            y: density matrix as (n,n) Array representing the state at time :math:`t`.
        Returns:
            RHS of Lindblad equation
            .. math::
                -i[H,y] + \sum_j\gamma_j(t)(L_j y L_j^\dagger - (1/2) * {L_j^\daggerL_j,y}).
        Raises:
            QiskitError: If operator collection is underspecified.
        """

        hamiltonian_matrix = None
        if self._static_hamiltonian is not None or self._hamiltonian_operators is not None:
            hamiltonian_matrix = -1j * self.evaluate_hamiltonian(ham_sig_vals)  # B matrix

        # if dissipators present (includes both hamiltonian is None and is not None)
        if self._dissipator_operators is not None or self._static_dissipators is not None:

            # A matrix
            if self._static_dissipators is None:
                dissipators_matrix = jsparse_linear_combo(dis_sig_vals, self._dissipator_products)
            elif self._dissipator_operators is None:
                dissipators_matrix = self._static_dissipators_product_sum
            else:
                dissipators_matrix = self._static_dissipators_product_sum + jsparse_linear_combo(
                    dis_sig_vals, self._dissipator_products
                )

            if hamiltonian_matrix is not None:
                left_mult_contribution = jsparse_matmul(hamiltonian_matrix + dissipators_matrix, y)
                right_mult_contribution = jsparse_matmul(
                    y, dissipators_matrix + (-1 * hamiltonian_matrix)
                )
            else:
                left_mult_contribution = jsparse_matmul(dissipators_matrix, y)
                right_mult_contribution = jsparse_matmul(y, dissipators_matrix)

            if len(y.shape) == 3:
                # Must do array broadcasting and transposition to ensure vectorization works
                y = jnp.broadcast_to(y, (1, y.shape[0], y.shape[1], y.shape[2])).transpose(
                    [1, 0, 2, 3]
                )
            if self._static_dissipators is None:
                both_mult_contribution = jnp.tensordot(
                    dis_sig_vals,
                    jsparse_triple_product(
                        self._dissipator_operators, y, self._dissipator_operators_adj
                    ),
                    axes=(-1, -3),
                )
            elif self._dissipator_operators is None:
                both_mult_contribution = jnp.sum(
                    jsparse_triple_product(
                        self._static_dissipators, y, self._static_dissipators_adj
                    ),
                    axis=-3,
                )
            else:
                both_mult_contribution = jnp.sum(
                    jsparse_triple_product(
                        self._static_dissipators, y, self._static_dissipators_adj
                    ),
                    axis=-3,
                ) + jnp.tensordot(
                    dis_sig_vals,
                    jsparse_triple_product(
                        self._dissipator_operators, y, self._dissipator_operators_adj
                    ),
                    axes=(-1, -3),
                )

            out = left_mult_contribution + right_mult_contribution + both_mult_contribution

            return out
        # if just hamiltonian
        elif hamiltonian_matrix is not None:
            return jsparse_matmul(hamiltonian_matrix, y) - jsparse_matmul(y, hamiltonian_matrix)
        else:
            raise QiskitError(
                """JAXSparseLindbladCollection with None for static_hamiltonian,
                                 hamiltonian_operators, static_dissipators, and
                                 dissipator_operators, cannot evaluate rhs."""
            )


class BaseVectorizedLindbladCollection(BaseLindbladOperatorCollection, BaseOperatorCollection):
    """Base class for Vectorized Lindblad collections.

    The vectorized Lindblad equation represents the Lindblad master equation in the structure
    of a linear matrix differential equation in standard form. Hence, this class inherits
    from both ``BaseLindbladOperatorCollection`` and ``BaseOperatorCollection``.

    This class manages the general property handling of converting operators in a Lindblad
    collection to the correct type, constructing vectorized versions, and combining for use in a
    BaseOperatorCollection. Requires implementation of:

        - ``convert_to_internal_type``: Convert operators to the required internal type,
          e.g. csr or Array.
        - ``evaluation_class``: Class property that returns the subclass of BaseOperatorCollection
          to be used when evaluating the model, e.g. DenseOperatorCollection or
          SparseOperatorCollection.
    """

    def __init__(
        self,
        static_hamiltonian: Optional[Array] = None,
        hamiltonian_operators: Optional[Array] = None,
        static_dissipators: Optional[Array] = None,
        dissipator_operators: Optional[Array] = None,
    ):
        r"""Initialize collection.

        Args:
            static_hamiltonian: Constant term :math:`H_d` to be added to the Hamiltonian of the
                                system.
            hamiltonian_operators: Specifies breakdown of Hamiltonian
                as :math:`H(t) = \sum_j s(t) H_j+H_d` by specifying H_j. (k,n,n) array.
            static_dissipators: Dissipator terms with coefficient 1.
            dissipator_operators: the terms :math:`L_j` in Lindblad equation. (m,n,n) array.
        """
        self._static_hamiltonian = None
        self._hamiltonian_operators = None
        self._static_dissipators = None
        self._dissipator_operators = None
        self._static_operator = None
        self._operators = None
        super().__init__(
            static_hamiltonian=static_hamiltonian,
            hamiltonian_operators=hamiltonian_operators,
            static_dissipators=static_dissipators,
            dissipator_operators=dissipator_operators,
        )

    @abstractmethod
    def convert_to_internal_type(self, obj: any) -> any:
        """Convert either a single operator or a list of operators to an internal representation."""

    @property
    @abstractmethod
    def evaluation_class(self) -> BaseOperatorCollection:
        """Class used for evaluating the vectorized model or RHS."""

    @property
    def static_hamiltonian(self) -> Union[Array, csr_matrix]:
        """Returns static part of operator collection."""
        return self._static_hamiltonian

    @static_hamiltonian.setter
    def static_hamiltonian(self, new_static_hamiltonian: Optional[Union[Array, csr_matrix]] = None):
        """Sets static_operator term."""
        self._static_hamiltonian = self.convert_to_internal_type(new_static_hamiltonian)
        if self._static_hamiltonian is not None:
            self._vec_static_hamiltonian = vec_commutator(self._static_hamiltonian)

        self.concatenate_static_operators()

    @property
    def hamiltonian_operators(self) -> Array:
        return self._hamiltonian_operators

    @hamiltonian_operators.setter
    def hamiltonian_operators(
        self, new_hamiltonian_operators: Optional[Union[Array, csr_matrix]] = None
    ):
        self._hamiltonian_operators = self.convert_to_internal_type(new_hamiltonian_operators)
        if self._hamiltonian_operators is not None:
            self._vec_hamiltonian_operators = vec_commutator(self._hamiltonian_operators)

        self.concatenate_operators()

    @property
    def static_dissipators(self) -> Union[Array, List[csr_matrix]]:
        return self._static_dissipators

    @static_dissipators.setter
    def static_dissipators(
        self, new_static_dissipators: Optional[Union[Array, List[csr_matrix]]] = None
    ):
        self._static_dissipators = self.convert_to_internal_type(new_static_dissipators)
        if self._static_dissipators is not None:
            self._vec_static_dissipators_sum = np.sum(
                vec_dissipator(self._static_dissipators), axis=0
            )

        self.concatenate_static_operators()

    @property
    def dissipator_operators(self) -> Union[Array, List[csr_matrix]]:
        return self._dissipator_operators

    @dissipator_operators.setter
    def dissipator_operators(
        self, new_dissipator_operators: Optional[Union[Array, List[csr_matrix]]] = None
    ):
        self._dissipator_operators = self.convert_to_internal_type(new_dissipator_operators)
        if self._dissipator_operators is not None:
            self._vec_dissipator_operators = vec_dissipator(self._dissipator_operators)

        self.concatenate_operators()

    def concatenate_static_operators(self):
        """Concatenate static hamiltonian and static dissipators."""
        if self._static_hamiltonian is not None and self._static_dissipators is not None:
            self._static_operator = self._vec_static_hamiltonian + self._vec_static_dissipators_sum
        elif self._static_hamiltonian is None and self._static_dissipators is not None:
            self._static_operator = self._vec_static_dissipators_sum
        elif self._static_hamiltonian is not None and self._static_dissipators is None:
            self._static_operator = self._vec_static_hamiltonian
        else:
            self._static_operator = None

    def concatenate_operators(self):
        """Concatenate hamiltonian operators and dissipator operators."""
        if self._hamiltonian_operators is not None and self._dissipator_operators is not None:
            self._operators = np.append(
                self._vec_hamiltonian_operators, self._vec_dissipator_operators, axis=0
            )
        elif self._hamiltonian_operators is not None and self._dissipator_operators is None:
            self._operators = self._vec_hamiltonian_operators
        elif self._hamiltonian_operators is None and self._dissipator_operators is not None:
            self._operators = self._vec_dissipator_operators
        else:
            self._operators = None

    def concatenate_signals(
        self, ham_sig_vals: Union[None, Array], dis_sig_vals: Union[None, Array]
    ) -> Array:
        """Concatenate hamiltonian and linblad signals."""
        if self._hamiltonian_operators is not None and self._dissipator_operators is not None:
            return np.append(ham_sig_vals, dis_sig_vals, axis=-1)
        if self._hamiltonian_operators is not None and self._dissipator_operators is None:
            return ham_sig_vals
        if self._hamiltonian_operators is None and self._dissipator_operators is not None:
            return dis_sig_vals

        return None

    def evaluate(self, ham_sig_vals: Union[None, Array], dis_sig_vals: Union[None, Array]) -> Array:
        """Evaluate the model."""
        signal_values = self.concatenate_signals(ham_sig_vals, dis_sig_vals)
        return self.evaluation_class.evaluate(self, signal_values)

    def evaluate_rhs(self, ham_sig_vals: Array, dis_sig_vals: Array, y: Array) -> Array:
        r"""Evaluates the RHS of the Lindblad equation using
        vectorized maps.

        Args:
            ham_sig_vals: hamiltonian signal coefficients.
            dis_sig_vals: dissipator signal coefficients.
                If none involved, pass None.
            y: Density matrix represented as a vector using column-stacking
                convention.
        Returns:
            Vectorized RHS of Lindblad equation :math:`\dot{\rho}` in column-stacking
                convention.
        """
        return self.evaluate(ham_sig_vals, dis_sig_vals) @ y


class DenseVectorizedLindbladCollection(
    BaseVectorizedLindbladCollection, DenseLindbladCollection, DenseOperatorCollection
):
    r"""Vectorized version of DenseLindbladCollection.

    Utilizes BaseVectorizedLindbladCollection for property handling, DenseLindbladCollection
    for evaluate_hamiltonian, and DenseOperatorCollection for operator property handling.
    """

    def convert_to_internal_type(self, obj: any) -> Array:
        return to_array(obj)

    @property
    def evaluation_class(self):
        return DenseOperatorCollection


class SparseVectorizedLindbladCollection(
    BaseVectorizedLindbladCollection, SparseLindbladCollection, SparseOperatorCollection
):
    r"""Vectorized version of SparseLindbladCollection.

    Utilizes BaseVectorizedLindbladCollection for property handling, SparseLindbladCollection
    for evaluate_hamiltonian, and SparseOperatorCollection for static_operator and operator
    property handling.
    """

    def convert_to_internal_type(self, obj: any) -> Union[csr_matrix, List[csr_matrix]]:
        if obj is None:
            return None

        obj = to_csr(obj)
        if issparse(obj):
            return np.round(obj, decimals=self._decimals)
        else:
            return [np.round(sub_obj, decimals=self._decimals) for sub_obj in obj]

    @property
    def evaluation_class(self):
        return SparseOperatorCollection


class JAXSparseVectorizedLindbladCollection(
    BaseVectorizedLindbladCollection, JAXSparseLindbladCollection, JAXSparseOperatorCollection
):
    r"""Vectorized version of JAXSparseLindbladCollection.

    Utilizes BaseVectorizedLindbladCollection for property handling, JAXSparseLindbladCollection
    for evaluate_hamiltonian, and JAXSparseOperatorCollection for static_operator and operator
    property handling.
    """

    def convert_to_internal_type(self, obj: any) -> "BCOO":
        return to_BCOO(obj)

    @property
    def evaluation_class(self):
        return JAXSparseOperatorCollection

    def concatenate_static_operators(self):
        """Override base class to convert to BCOO again at the end. The vectorization operations
        are not implemented for BCOO type, so they automatically get converted to Arrays,
        and hence need to be converted back.
        """
        super().concatenate_static_operators()
        self._static_operator = self.convert_to_internal_type(self._static_operator)

    def concatenate_operators(self):
        """Override base class to convert to BCOO again at the end. The vectorization operations
        are not implemented for BCOO type, so they automatically get converted to Arrays,
        and hence need to be converted back.
        """
        super().concatenate_operators()
        self._operators = self.convert_to_internal_type(self._operators)

    @wrap
    def evaluate_rhs(self, ham_sig_vals: Array, dis_sig_vals: Array, y: Array) -> Array:
        return jsparse_matmul(self.evaluate(ham_sig_vals, dis_sig_vals), y)


def package_density_matrices(y: Array) -> Array:
    """Sends a (k,n,n) Array y of density matrices to a
    (k,1) Array of dtype object, where entry [j,0] is
    y[j]. Formally avoids For loops through vectorization.
    Args:
        y: (k,n,n) Array.
    Returns:
        Array with dtype object."""
    # As written here, only works for (n,n) Arrays
    obj_arr = np.empty(shape=(1), dtype="O")
    obj_arr[0] = y
    return obj_arr


# Using vectorization with signature, works on (k,n,n) Arrays -> (k,1) Array
package_density_matrices = np.vectorize(package_density_matrices, signature="(n,n)->(1)")


def unpackage_density_matrices(y: Array) -> Array:
    """Inverse function of package_density_matrices,
    Much slower than packaging. Avoid using unless
    absolutely needed (as in case of passing multiple
    density matrices to SparseLindbladCollection.evaluate_rhs)."""
    return y[0]


unpackage_density_matrices = np.vectorize(unpackage_density_matrices, signature="(1)->(n,n)")



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

r"""
=========================================================
Perturbation Theory (:mod:`qiskit_dynamics.perturbation`)
=========================================================

.. currentmodule:: qiskit_dynamics.perturbation

This module contains tools for numerically computing and utilizing perturbation theory terms.
Perturbation theory is an advanced topic; a brief review of the concepts and notation required
to understand the contents in this module are given in the
:ref:`Time-dependent perturbation theory and multi-variable
series expansions review <perturbation review>` discussion.

.. _td perturbation theory:

Time-dependent perturbation theory
==================================

The function :func:`.solve_lmde_perturbation` computes
Dyson series :footcite:`dyson_radiation_1949` and
Magnus expansion :footcite:`magnus_exponential_1954,blanes_magnus_2009` terms
in a multi-variable setting via algorithms in :footcite:`puzzuoli_sensitivity_2022`.
It can also be used to compute Dyson-like integrals using the algorithm in
:footcite:`haas_engineering_2019`. Results are returned in either a :class:`PowerSeriesData`
or :class:`DysonLikeData` class, which are data classes with functionality for indexing
and accessing specific perturbation terms. See the function documentation for further details.

Truncated power-series representation and multisets
===================================================

The class :class:`.ArrayPolynomial` represents an array-valued multivariable polynomial
(i.e. a truncated power series), and provides functionality for
both evaluating and transforming array-valued polynomials.

This module makes use of the `multiset package <https://pypi.org/project/multiset/>`_ for
indexing multi-variable power series. See the
:ref:`multiset and power series notation section <multiset power series>`
of the perturbation review for an explanation of this convention.


Perturbation module functions
=============================

.. autosummary::
    :toctree: ../stubs/

    solve_lmde_perturbation


Perturbation module classes
===========================

.. autosummary::
    :toctree: ../stubs/

    ArrayPolynomial
    PowerSeriesData
    DysonLikeData

.. footbibliography::
"""

from .array_polynomial import ArrayPolynomial
from .solve_lmde_perturbation import solve_lmde_perturbation
from .perturbation_data import PowerSeriesData, DysonLikeData



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
=================================================
Dispatch Module (:mod:`qiskit_dynamics.dispatch`)
=================================================

.. currentmodule:: qiskit_dynamics.dispatch

This module contains dispatch methods used by the
:class:`~qiskit_dynamics.array.Array` class.


Dispatch Functions
==================

.. autosummary::
    :toctree: ../stubs/

    asarray
    requires_backend
"""

# Import dispatch utilities
from .dispatch import (
    asarray,
    requires_backend,
)

# Register backends
from .backends import *



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Monkey patch qiskit.quantum_info to work easier with Array."""

from qiskit.quantum_info import (
    Statevector,
    DensityMatrix,
    Operator,
    SuperOp,
    Choi,
    PTM,
    Chi,
    Pauli,
    Clifford,
    SparsePauliOp,
)
from .array import Array

__all__ = []


def __qiskit_array__(self, dtype=None, backend=None):
    """Convert qi operator to an Array"""
    return Array(self.data, dtype=dtype, backend=backend)


def __to_matrix_array__(self, dtype=None, backend=None):
    """Convert object to array through to_matrix method"""
    return Array(self.to_matrix(), dtype=dtype, backend=backend)


# Monkey patch quantum info operators

Statevector.__qiskit_array__ = __qiskit_array__
DensityMatrix.__qiskit_array__ = __qiskit_array__
Operator.__qiskit_array__ = __qiskit_array__
SuperOp.__qiskit_array__ = __qiskit_array__
Choi.__qiskit_array__ = __qiskit_array__
Chi.__qiskit_array__ = __qiskit_array__
PTM.__qiskit_array__ = __qiskit_array__
Pauli.__qiskit_array__ = __to_matrix_array__
Clifford.__qiskit_array__ = __to_matrix_array__
SparsePauliOp.__qiskit_array__ = __to_matrix_array__



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Lindblad models module.
"""

from typing import Tuple, Union, List, Optional
from warnings import warn
from scipy.sparse.csr import csr_matrix

from qiskit import QiskitError
from qiskit.quantum_info.operators import Operator
from qiskit_dynamics.array import Array
from qiskit_dynamics.type_utils import to_numeric_matrix_type
from qiskit_dynamics.signals import Signal, SignalList
from .generator_model import (
    BaseGeneratorModel,
    transfer_static_operator_between_frames,
    transfer_operators_between_frames,
)
from .hamiltonian_model import HamiltonianModel, is_hermitian
from .operator_collections import (
    BaseLindbladOperatorCollection,
    DenseLindbladCollection,
    DenseVectorizedLindbladCollection,
    SparseLindbladCollection,
    JAXSparseLindbladCollection,
    SparseVectorizedLindbladCollection,
    JAXSparseVectorizedLindbladCollection,
)
from .rotating_frame import RotatingFrame

try:
    import jax
except ImportError:
    pass


class LindbladModel(BaseGeneratorModel):
    r"""A model of a quantum system in terms of the Lindblad master equation.

    The Lindblad master equation models the evolution of a density matrix according to:

    .. math::

        \dot{\rho}(t) = -i[H(t), \rho(t)] + \mathcal{D}_0(\rho(t)) + \mathcal{D}(t)(\rho(t)),

    where :math:`\mathcal{D}_0` is the static dissipator portion, and :math:`\mathcal{D}(t)` is the
    time-dependent dissipator portion of the equation, given by

    .. math::

        \mathcal{D}_0(\rho(t)) = \sum_j N_j \rho N_j^\dagger
                                      - \frac{1}{2} \{N_j^\dagger N_j, \rho\},

    and

    .. math::

        \mathcal{D}(t)(\rho(t)) = \sum_j \gamma_j(t) L_j \rho L_j^\dagger
                                  - \frac{1}{2} \{L_j^\dagger L_j, \rho\},

    respectively. In the above:

        - :math:`[\cdot, \cdot]` and :math:`\{\cdot, \cdot\}`
          respectively denote the matrix commutator and anti-commutator,
        - :math:`H(t)` denotes the Hamiltonian,
        - :math:`N_j` denotes the operators appearing in the static dissipator,
        - :math:`L_j` denotes the operators appearing in the time-dpendent portion of the
          dissipator, and
        - :math:`\gamma_j(t)` denotes the signal corresponding to the
          :math:`j^{th}` time-dependent dissipator operator.

    Instantiating an instance of :class:`~qiskit_dynamics.models.LindbladModel`
    requires specifying the above decomposition:

    .. code-block:: python

        lindblad_model = LindbladModel(
            static_hamiltonian=static_hamiltonian,
            hamiltonian_operators=hamiltonian_operators,
            hamiltonian_signals=hamiltonian_signals,
            static_dissipators=static_dissipators,
            dissipator_operators=dissipator_operators,
            dissipator_signals=dissipator_signals
        )

    where the arguments ``hamiltonian_operators``, ``hamiltonian_signals``, and
    ``static_hamiltonian`` are for the Hamiltonian decomposition as in
    :class:`~qiskit_dynamics.models.HamiltonianModel`,
    and the ``static_dissipators`` correspond to the :math:`N_j`, the ``dissipator_operators``
    to the :math:`L_j`, and the ``dissipator_signals`` the :math:`\gamma_j(t)`.
    """

    def __init__(
        self,
        static_hamiltonian: Optional[Union[Array, csr_matrix]] = None,
        hamiltonian_operators: Optional[Union[Array, List[csr_matrix]]] = None,
        hamiltonian_signals: Optional[Union[List[Signal], SignalList]] = None,
        static_dissipators: Optional[Union[Array, csr_matrix]] = None,
        dissipator_operators: Optional[Union[Array, List[csr_matrix]]] = None,
        dissipator_signals: Optional[Union[List[Signal], SignalList]] = None,
        rotating_frame: Optional[Union[Operator, Array, RotatingFrame]] = None,
        in_frame_basis: bool = False,
        evaluation_mode: Optional[str] = "dense",
        validate: bool = True,
    ):
        """Initialize.

        Args:
            static_hamiltonian: Constant term in Hamiltonian.
            hamiltonian_operators: List of operators in Hamiltonian with time-dependent
                                   coefficients.
            hamiltonian_signals: Time-dependent coefficients for hamiltonian_operators.
            static_dissipators: List of dissipators with coefficient 1.
            dissipator_operators: List of dissipator operators with time-dependent coefficients.
            dissipator_signals: Time-dependent coefficients for dissipator_operators.
            rotating_frame: Rotating frame in which calcualtions are to be done.
                            If provided, it is assumed that all operators were
                            already in the frame basis.
            in_frame_basis: Whether to represent the model in the basis in which the rotating
                            frame operator is diagonalized.
            evaluation_mode: Evaluation mode to use. Call ``help(LindbladModel.evaluation_mode)``
                             for more details.
            validate: If True check input hamiltonian_operators and static_hamiltonian are
                      Hermitian.

        Raises:
            QiskitError: If model insufficiently or incorrectly specified.
        """

        if (
            static_hamiltonian is None
            and hamiltonian_operators is None
            and static_dissipators is None
            and dissipator_operators is None
        ):
            raise QiskitError(
                f"{type(self).__name__} requires at least one of static_hamiltonian "
                "hamiltonian_operators, static_dissipators, or dissipator_operators "
                "to be specified at construction."
            )

        static_hamiltonian = to_numeric_matrix_type(static_hamiltonian)
        hamiltonian_operators = to_numeric_matrix_type(hamiltonian_operators)
        if validate:
            if (hamiltonian_operators is not None) and (not is_hermitian(hamiltonian_operators)):
                raise QiskitError("""LinbladModel hamiltonian_operators must be Hermitian.""")
            if (static_hamiltonian is not None) and (not is_hermitian(static_hamiltonian)):
                raise QiskitError("""LinbladModel static_hamiltonian must be Hermitian.""")

        self._operator_collection = construct_lindblad_operator_collection(
            evaluation_mode=evaluation_mode,
            static_hamiltonian=static_hamiltonian,
            hamiltonian_operators=hamiltonian_operators,
            static_dissipators=static_dissipators,
            dissipator_operators=dissipator_operators,
        )
        self._evaluation_mode = evaluation_mode
        self.vectorized_operators = "vectorized" in evaluation_mode

        self._rotating_frame = None
        self.rotating_frame = rotating_frame

        self._in_frame_basis = None
        self.in_frame_basis = in_frame_basis

        self.signals = (hamiltonian_signals, dissipator_signals)

    @classmethod
    def from_hamiltonian(
        cls,
        hamiltonian: HamiltonianModel,
        static_dissipators: Optional[Union[Array, csr_matrix]] = None,
        dissipator_operators: Optional[Union[Array, List[csr_matrix]]] = None,
        dissipator_signals: Optional[Union[List[Signal], SignalList]] = None,
        evaluation_mode: Optional[str] = None,
    ):
        """Construct from a :class:`HamiltonianModel`.

        Args:
            hamiltonian: The :class:`HamiltonianModel`.
            static_dissipators: List of dissipators with coefficient 1.
            dissipator_operators: List of dissipators with time-dependent coefficients.
            dissipator_signals: List time-dependent coefficients for dissipator_operators.
            evaluation_mode: Evaluation mode. Call ``help(LindbladModel.evaluation_mode)``
                for more information.

        Returns:
            LindbladModel: Linblad model from parameters.
        """

        if evaluation_mode is None:
            evaluation_mode = hamiltonian.evaluation_mode

        ham_copy = hamiltonian.copy()
        ham_copy.rotating_frame = None

        return cls(
            static_hamiltonian=ham_copy._get_static_operator(in_frame_basis=False),
            hamiltonian_operators=ham_copy._get_operators(in_frame_basis=False),
            hamiltonian_signals=ham_copy.signals,
            static_dissipators=static_dissipators,
            dissipator_operators=dissipator_operators,
            dissipator_signals=dissipator_signals,
            rotating_frame=hamiltonian.rotating_frame,
            in_frame_basis=hamiltonian.in_frame_basis,
            evaluation_mode=evaluation_mode,
        )

    @property
    def dim(self) -> int:
        if self._operator_collection.static_hamiltonian is not None:
            return self._operator_collection.static_hamiltonian.shape[-1]
        elif self._operator_collection.hamiltonian_operators is not None:
            return self._operator_collection.hamiltonian_operators[0].shape[-1]
        elif self._operator_collection.static_dissipators is not None:
            return self._operator_collection.static_dissipators[0].shape[-1]
        else:
            return self._operator_collection.dissipator_operators[0].shape[-1]

    @property
    def signals(self) -> Tuple[SignalList]:
        """Gets the Model's Signals.

        Returns:
            Tuple[] with 0th entry storing the Hamiltonian signals
            and the 1st entry storing the dissipator signals.
        """
        return (self._hamiltonian_signals, self._dissipator_signals)

    @signals.setter
    def signals(self, new_signals: Tuple[List[Signal]]):
        """Set signals.

        Raises:
            QiskitError: If signals incompatible with operator structure.
        """
        hamiltonian_signals, dissipator_signals = new_signals

        # set Hamiltonian signals
        if hamiltonian_signals is None:
            self._hamiltonian_signals = None
        elif hamiltonian_signals is not None and self.hamiltonian_operators is None:
            raise QiskitError("Hamiltonian signals must be None if hamiltonian_operators is None.")
        else:
            # if signals is a list, instantiate a SignalList
            if isinstance(hamiltonian_signals, list):
                hamiltonian_signals = SignalList(hamiltonian_signals)

            # if it isn't a SignalList by now, raise an error
            if not isinstance(hamiltonian_signals, SignalList):
                raise QiskitError("Hamiltonian signals specified in unaccepted format.")

            # verify signal length is same as operators
            if isinstance(self.hamiltonian_operators, list):
                len_hamiltonian_operators = len(self.hamiltonian_operators)
            else:
                len_hamiltonian_operators = self.hamiltonian_operators.shape[0]
            if len(hamiltonian_signals) != len_hamiltonian_operators:
                raise QiskitError(
                    "Hamiltonian signals need to have the same length as Hamiltonian operators."
                )

            self._hamiltonian_signals = hamiltonian_signals

        # set dissipator signals
        if dissipator_signals is None:
            self._dissipator_signals = None
        elif dissipator_signals is not None and self.dissipator_operators is None:
            raise QiskitError("Dissipator signals must be None if dissipator_operators is None.")
        else:
            # if signals is a list, instantiate a SignalList
            if isinstance(dissipator_signals, list):
                dissipator_signals = SignalList(dissipator_signals)

            # if it isn't a SignalList by now, raise an error
            if not isinstance(dissipator_signals, SignalList):
                raise QiskitError("Dissipator signals specified in unaccepted format.")

            # verify signal length is same as operators
            if isinstance(self.dissipator_operators, list):
                len_dissipator_operators = len(self.dissipator_operators)
            else:
                len_dissipator_operators = self.dissipator_operators.shape[0]
            if len(dissipator_signals) != len_dissipator_operators:
                raise QiskitError(
                    "Dissipator signals need to have the same length as dissipator operators."
                )

            self._dissipator_signals = dissipator_signals

    @property
    def in_frame_basis(self) -> bool:
        """Whether to represent the model in the basis in which the frame operator
        is diagonalized.
        """
        return self._in_frame_basis

    @in_frame_basis.setter
    def in_frame_basis(self, in_frame_basis: bool):
        self._in_frame_basis = in_frame_basis

    @property
    def static_hamiltonian(self) -> Array:
        """Get the static Hamiltonian term."""
        return self._get_static_hamiltonian(in_frame_basis=self._in_frame_basis)

    @static_hamiltonian.setter
    def static_hamiltonian(self, static_hamiltonian: Array):
        """Set the static Hamiltonian term."""
        self._set_static_hamiltonian(
            new_static_hamiltonian=static_hamiltonian, operator_in_frame_basis=self._in_frame_basis
        )

    @property
    def hamiltonian_operators(self) -> Array:
        """Get the Hamiltonian operators."""
        return self._get_hamiltonian_operators(in_frame_basis=self._in_frame_basis)

    @property
    def static_dissipators(self) -> Array:
        """Get the static dissipators."""
        return self._get_static_dissipators(in_frame_basis=self._in_frame_basis)

    @property
    def dissipator_operators(self) -> Array:
        """Get the dissipator operators."""
        return self._get_dissipator_operators(in_frame_basis=self._in_frame_basis)

    def _get_static_hamiltonian(self, in_frame_basis: Optional[bool] = False) -> Array:
        """Get the constant hamiltonian term.

        Args:
            in_frame_basis: Flag for whether the returned static_operator should be
            in the basis in which the frame is diagonal.
        Returns:
            The static operator term.
        """
        op = self._operator_collection.static_hamiltonian
        if not in_frame_basis and self.rotating_frame is not None:
            return self.rotating_frame.operator_out_of_frame_basis(op)
        else:
            return op

    def _set_static_hamiltonian(
        self,
        new_static_hamiltonian: Array,
        operator_in_frame_basis: Optional[bool] = False,
    ):
        """Set the constant Hamiltonian term.
        Note that if the model has a rotating frame this will override
        any contributions to the static term due to the frame transformation.

        Args:
            new_static_hamiltonian: The static operator operator.
            operator_in_frame_basis: Whether `new_static_operator` is already in the rotating
            frame basis.
        """
        if new_static_hamiltonian is None:
            self._operator_collection.static_hamiltonian = None
        else:
            if not operator_in_frame_basis and self.rotating_frame is not None:
                new_static_hamiltonian = self.rotating_frame.operator_into_frame_basis(
                    new_static_hamiltonian
                )

            self._operator_collection.static_hamiltonian = new_static_hamiltonian

    def _get_hamiltonian_operators(self, in_frame_basis: Optional[bool] = False) -> Tuple[Array]:
        """Get the Hamiltonian operators, either in the frame basis or not.

        Args:
            in_frame_basis: Whether to return in frame basis or not.
        Returns:
            Hamiltonian operators.
        """
        ham_ops = self._operator_collection.hamiltonian_operators
        if not in_frame_basis and self.rotating_frame is not None:
            ham_ops = self.rotating_frame.operator_out_of_frame_basis(ham_ops)

        return ham_ops

    def _get_static_dissipators(self, in_frame_basis: Optional[bool] = False) -> Tuple[Array]:
        """Get the static dissipators, either in the frame basis or not.

        Args:
            in_frame_basis: Whether to return in frame basis or not.
        Returns:
            Dissipator operators.
        """
        diss_ops = self._operator_collection.static_dissipators
        if not in_frame_basis and self.rotating_frame is not None:
            diss_ops = self.rotating_frame.operator_out_of_frame_basis(diss_ops)

        return diss_ops

    def _get_dissipator_operators(self, in_frame_basis: Optional[bool] = False) -> Tuple[Array]:
        """Get the Dissipator operators, either in the frame basis or not.

        Args:
            in_frame_basis: Whether to return in frame basis or not.
        Returns:
            Dissipator operators.
        """
        diss_ops = self._operator_collection.dissipator_operators
        if not in_frame_basis and self.rotating_frame is not None:
            diss_ops = self.rotating_frame.operator_out_of_frame_basis(diss_ops)

        return diss_ops

    @property
    def evaluation_mode(self) -> str:
        """Numerical evaluation mode of the model.

        Available options:

         * 'dense': Stores Hamiltonian and dissipator terms as dense Array types.
         * 'dense_vectorized': Stores the Hamiltonian and dissipator
            terms as :math:`(dim^2,dim^2)` matrices that acts on a vectorized
            density matrix by left-multiplication. Allows for direct evaluate generator.
         * 'sparse': Like dense, but matrices stored in sparse format. If the default Array backend
            is JAX, uses JAX BCOO sparse arrays, otherwise uses scipy
            :class:`csr_matrix` sparse type.
         * `sparse_vectorized`: Like dense_vectorized, but matrices stored in sparse format.
            If the default Array backend is JAX, uses JAX BCOO sparse arrays, otherwise uses scipy
            :class:`csr_matrix` sparse type. Note that
            JAX sparse mode is only recommended for use on CPU.
        """
        return self._evaluation_mode

    @evaluation_mode.setter
    def evaluation_mode(self, new_mode: str):
        """Sets evaluation mode.

        Args:
            new_mode: String specifying new mode. Available options
                      are 'dense', 'sparse', 'dense_vectorized', and 'sparse_vectorized'.
                      See property doc string for details.

        Raises:
            NotImplementedError: if new_mode is not one of the above
            supported evaluation modes.
        """
        if new_mode != self._evaluation_mode:
            self._operator_collection = construct_lindblad_operator_collection(
                evaluation_mode=new_mode,
                static_hamiltonian=self._operator_collection.static_hamiltonian,
                hamiltonian_operators=self._operator_collection.hamiltonian_operators,
                static_dissipators=self._operator_collection.static_dissipators,
                dissipator_operators=self._operator_collection.dissipator_operators,
            )

            self.vectorized_operators = "vectorized" in new_mode
            self._evaluation_mode = new_mode

    @property
    def rotating_frame(self):
        return self._rotating_frame

    @rotating_frame.setter
    def rotating_frame(self, rotating_frame: Union[Operator, Array, RotatingFrame]):
        new_frame = RotatingFrame(rotating_frame)

        # convert static hamiltonian to new frame setup
        static_ham = self._get_static_hamiltonian(in_frame_basis=True)
        if static_ham is not None:
            static_ham = -1j * static_ham

        new_static_hamiltonian = transfer_static_operator_between_frames(
            static_ham,
            new_frame=new_frame,
            old_frame=self.rotating_frame,
        )

        if new_static_hamiltonian is not None:
            new_static_hamiltonian = 1j * new_static_hamiltonian

        # convert hamiltonian operators and dissipator operators
        ham_ops = self._get_hamiltonian_operators(in_frame_basis=True)
        static_diss_ops = self._get_static_dissipators(in_frame_basis=True)
        diss_ops = self._get_dissipator_operators(in_frame_basis=True)

        new_hamiltonian_operators = transfer_operators_between_frames(
            ham_ops,
            new_frame=new_frame,
            old_frame=self.rotating_frame,
        )
        new_static_dissipators = transfer_operators_between_frames(
            static_diss_ops,
            new_frame=new_frame,
            old_frame=self.rotating_frame,
        )
        new_dissipator_operators = transfer_operators_between_frames(
            diss_ops,
            new_frame=new_frame,
            old_frame=self.rotating_frame,
        )

        self._rotating_frame = new_frame

        self._operator_collection = construct_lindblad_operator_collection(
            evaluation_mode=self.evaluation_mode,
            static_hamiltonian=new_static_hamiltonian,
            hamiltonian_operators=new_hamiltonian_operators,
            static_dissipators=new_static_dissipators,
            dissipator_operators=new_dissipator_operators,
        )

    def evaluate_hamiltonian(self, time: float) -> Array:
        """Evaluates Hamiltonian matrix at a given time.

        Args:
            time: The time at which to evaluate the hamiltonian.
        Returns:
            Array: Hamiltonian matrix."""

        hamiltonian_sig_vals = None
        if self._hamiltonian_signals is not None:
            hamiltonian_sig_vals = self._hamiltonian_signals(time)

        ham = self._operator_collection.evaluate_hamiltonian(hamiltonian_sig_vals)
        if self.rotating_frame.frame_diag is not None:
            ham = self.rotating_frame.operator_into_frame(
                time,
                ham,
                operator_in_frame_basis=True,
                return_in_frame_basis=self._in_frame_basis,
                vectorized_operators=self.vectorized_operators,
            )

        return ham

    def evaluate(self, time: float) -> Array:
        hamiltonian_sig_vals = None
        if self._hamiltonian_signals is not None:
            hamiltonian_sig_vals = self._hamiltonian_signals(time)
        elif self._operator_collection.hamiltonian_operators is not None:
            raise QiskitError(
                self.__class__.__name__
                + """ with non-empty hamiltonian operators cannot be evaluated without
                hamiltonian signals."""
            )

        dissipator_sig_vals = None
        if self._dissipator_signals is not None:
            dissipator_sig_vals = self._dissipator_signals(time)
        elif self._operator_collection.dissipator_operators is not None:
            raise QiskitError(
                self.__class__.__name__
                + """ with non-empty dissipator operators cannot be evaluated without
                dissipator signals."""
            )

        if self.vectorized_operators:
            out = self._operator_collection.evaluate(hamiltonian_sig_vals, dissipator_sig_vals)
            return self.rotating_frame.vectorized_map_into_frame(
                time, out, operator_in_frame_basis=True, return_in_frame_basis=self._in_frame_basis
            )
        else:
            raise NotImplementedError(
                "Non-vectorized Lindblad models cannot be represented without a given state."
            )

    def evaluate_rhs(self, time: Union[float, int], y: Array) -> Array:
        """Evaluates the Lindblad model at a given time.

        Args:
            time: time at which the model should be evaluated.
            y: Density matrix as an (n,n) Array if not using a
               vectorized evaluation_mode or an (n^2) Array if
               using vectorized evaluation.
        Returns:
            Array: Either the evaluated generator or the state.
        Raises:
            QiskitError: If signals not sufficiently specified.
        """

        hamiltonian_sig_vals = None
        if self._hamiltonian_signals is not None:
            hamiltonian_sig_vals = self._hamiltonian_signals(time)
        elif self._operator_collection.hamiltonian_operators is not None:
            raise QiskitError(
                self.__class__.__name__
                + """ with non-empty hamiltonian operators cannot be evaluated without
                hamiltonian signals."""
            )

        dissipator_sig_vals = None
        if self._dissipator_signals is not None:
            dissipator_sig_vals = self._dissipator_signals(time)
        elif self._operator_collection.dissipator_operators is not None:
            raise QiskitError(
                self.__class__.__name__
                + """ with non-empty dissipator operators cannot be evaluated without
                dissipator signals."""
            )

        if self.rotating_frame.frame_diag is not None:

            # Take y out of the frame, but keep in the frame basis
            rhs = self.rotating_frame.operator_out_of_frame(
                time,
                y,
                operator_in_frame_basis=self._in_frame_basis,
                return_in_frame_basis=True,
                vectorized_operators=self.vectorized_operators,
            )

            rhs = self._operator_collection.evaluate_rhs(
                hamiltonian_sig_vals, dissipator_sig_vals, rhs
            )

            # Put rhs back into the frame, potentially converting its basis.
            rhs = self.rotating_frame.operator_into_frame(
                time,
                rhs,
                operator_in_frame_basis=True,
                return_in_frame_basis=self._in_frame_basis,
                vectorized_operators=self.vectorized_operators,
            )

        else:
            rhs = self._operator_collection.evaluate_rhs(
                hamiltonian_sig_vals, dissipator_sig_vals, y
            )

        return rhs


def construct_lindblad_operator_collection(
    evaluation_mode: str,
    static_hamiltonian: Union[None, Array, csr_matrix],
    hamiltonian_operators: Union[None, Array, List[csr_matrix]],
    static_dissipators: Union[None, Array, csr_matrix],
    dissipator_operators: Union[None, Array, List[csr_matrix]],
) -> BaseLindbladOperatorCollection:
    """Construct Lindblad operator collection.

    Args:
        evaluation_mode: String specifying new mode. Available options
                         are 'dense', 'sparse', 'dense_vectorized', and 'sparse_vectorized'.
                         See property doc string for details.
        static_hamiltonian: Constant part of the Hamiltonian.
        hamiltonian_operators: Operators in Hamiltonian with time-dependent coefficients.
        static_dissipators: Dissipation operators with coefficient 1.
        dissipator_operators: Dissipation operators with variable coefficients.
    Returns:
        BaseLindbladOperatorCollection: Right-hand side evaluation object.
    Raises:
        NotImplementedError: if evaluation_mode is not one of the above
        supported evaluation modes.
    """

    # raise warning if sparse mode set with JAX not on cpu
    if (
        Array.default_backend() == "jax"
        and "sparse" in evaluation_mode
        and jax.default_backend() != "cpu"
    ):
        warn(
            """Using sparse mode with JAX is primarily recommended for use on CPU.""",
            stacklevel=2,
        )

    if evaluation_mode == "dense":
        CollectionClass = DenseLindbladCollection
    elif evaluation_mode == "sparse":
        if Array.default_backend() == "jax":
            CollectionClass = JAXSparseLindbladCollection
        else:
            CollectionClass = SparseLindbladCollection
    elif evaluation_mode == "dense_vectorized":
        CollectionClass = DenseVectorizedLindbladCollection
    elif evaluation_mode == "sparse_vectorized":
        if Array.default_backend() == "jax":
            CollectionClass = JAXSparseVectorizedLindbladCollection
        else:
            CollectionClass = SparseVectorizedLindbladCollection
    else:
        raise NotImplementedError(
            "Evaluation mode '"
            + str(evaluation_mode)
            + "' is not supported. Call help(LindbladModel.evaluation_mode) for available options."
        )

    return CollectionClass(
        static_hamiltonian=static_hamiltonian,
        hamiltonian_operators=hamiltonian_operators,
        static_dissipators=static_dissipators,
        dissipator_operators=dissipator_operators,
    )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Module for rotating frame handling classes.
"""

from typing import Union, List, Optional
import numpy as np
from scipy.sparse import issparse, csr_matrix

from qiskit import QiskitError
from qiskit.quantum_info.operators import Operator
from qiskit.quantum_info.operators.predicates import is_hermitian_matrix
from qiskit_dynamics.array import Array
from qiskit_dynamics.type_utils import to_array, to_BCOO, to_numeric_matrix_type

try:
    import jax.numpy as jnp
    from jax.experimental import sparse as jsparse

    jsparse_matmul = jsparse.sparsify(jnp.matmul)
except ImportError:
    pass


class RotatingFrame:
    r"""Class for representing a rotation frame transformation.

    This class provides functionality for transforming various objects into or out-of
    a rotating frame specified by an anti-Hermitian operator :math:`F = -iH`.
    For example:

        * Bringing a "state" into/out of the frame:
          :math:`t, y \mapsto e^{\mp tF}y`
        * Bringing an "operator" into/out of the frame:
          :math:`t, A \mapsto e^{\mp tF}Ae^{\pm tF}`
        * Bringing a generator for a BMDE into/out of the frame:
          :math:`t, G \mapsto e^{\mp tF}Ge^{\pm tF} - F`

    This class also contains functions for bringing states/operators into/out of
    the basis in which :math:`F` is diagonalized, which we refer to as the
    "frame basis". All previously mentioned functions also include optional
    arguments specifying whether the input/output are meant to be in the
    frame basis.

    .. note::
        :class:`~qiskit_dynamics.models.RotatingFrame` can be instantiated
        with a 1d array, which is understood to correspond to the diagonal entries
        of a diagonal :math:`H` or :math:`F = -i H`.

    """

    def __init__(
        self,
        frame_operator: Union[Array, Operator],
        atol: float = 1e-10,
        rtol: float = 1e-10,
    ):
        """Initialize with a frame operator.

        Args:
            frame_operator: The frame operator, must be either
                            Hermitian or anti-Hermitian.
            atol: Absolute tolerance when verifying that the frame_operator is
                  Hermitian or anti-Hermitian.
            rtol: Relative tolerance when verifying that the frame_operator is
                  Hermitian or anti-Hermitian.
        """
        if isinstance(frame_operator, RotatingFrame):
            frame_operator = frame_operator.frame_operator

        self._frame_operator = frame_operator
        frame_operator = to_array(frame_operator)

        if frame_operator is None:
            self._dim = None
            self._frame_diag = None
            self._frame_basis = None
            self._frame_basis_adjoint = None
        # if frame_operator is a 1d array, assume already diagonalized
        elif frame_operator.ndim == 1:

            # verify Hermitian or anti-Hermitian
            # if Hermitian convert to anti-Hermitian
            frame_operator = _is_herm_or_anti_herm(frame_operator, atol=atol, rtol=rtol)

            self._frame_diag = Array(frame_operator)
            self._frame_basis = None
            self._frame_basis_adjoint = None
            self._dim = len(self._frame_diag)
        # if not, diagonalize it
        else:

            # verify Hermitian or anti-Hermitian
            # if Hermitian convert to anti-Hermitian
            frame_operator = _is_herm_or_anti_herm(frame_operator, atol=atol, rtol=rtol)

            # diagonalize with eigh, utilizing assumption of anti-hermiticity
            frame_diag, frame_basis = np.linalg.eigh(1j * frame_operator)

            self._frame_diag = Array(-1j * frame_diag)
            self._frame_basis = Array(frame_basis)
            self._frame_basis_adjoint = frame_basis.conj().transpose()
            self._dim = len(self._frame_diag)

        # lazily evaluate change-of-basis matrices for vectorized operators.
        self._vectorized_frame_basis = None
        self._vectorized_frame_basis_adjoint = None

    @property
    def dim(self) -> int:
        """The dimension of the frame."""
        return self._dim

    @property
    def frame_operator(self) -> Array:
        """The original frame operator."""
        return self._frame_operator

    @property
    def frame_diag(self) -> Array:
        """Diagonal of the frame operator."""
        return self._frame_diag

    @property
    def frame_basis(self) -> Array:
        """Array containing diagonalizing unitary."""
        return self._frame_basis

    @property
    def frame_basis_adjoint(self) -> Array:
        """Adjoint of the diagonalizing unitary."""
        return self._frame_basis_adjoint

    def state_into_frame_basis(self, y: Array) -> Array:
        r"""Take a state into the frame basis, i.e. return
        ``self.frame_basis_adjoint @ y``.

        Args:
            y: The state.
        Returns:
            Array: The state in the frame basis.
        """
        y = to_numeric_matrix_type(y)
        if self.frame_basis_adjoint is None:
            return y

        return self.frame_basis_adjoint @ y

    def state_out_of_frame_basis(self, y: Array) -> Array:
        r"""Take a state out of the frame basis, i.e.
        ``return self.frame_basis @ y``.

        Args:
            y: The state.
        Returns:
            Array: The state in the frame basis.
        """
        y = to_numeric_matrix_type(y)
        if self.frame_basis is None:
            return y

        return self.frame_basis @ y

    def operator_into_frame_basis(
        self,
        op: Union[Operator, List[Operator], Array, csr_matrix, None],
        convert_type: bool = True,
    ) -> Array:
        r"""Take an operator into the frame basis, i.e. return
        ``self.frame_basis_adjoint @ A @ self.frame_basis``

        Args:
            op: The operator or array of operators.
            convert_type: Whether or not to initially convert ``op`` into an expected type.
                          Should only be set to ``False`` in situations in which it is gauranteed
                          that ``op`` is a handled input type.
        Returns:
            Array: The operator in the frame basis.
        """
        if convert_type:
            op = to_numeric_matrix_type(op)

        if self.frame_basis is None or op is None:
            return op

        if isinstance(op, list):
            return [self.operator_into_frame_basis(x, convert_type=False) for x in op]
        elif type(op).__name__ == "BCOO":
            return self.frame_basis_adjoint @ jsparse_matmul(op, self.frame_basis.data)
        else:
            # parentheses are necessary for sparse op evaluation
            return self.frame_basis_adjoint @ (op @ self.frame_basis)

    def operator_out_of_frame_basis(
        self,
        op: Union[Operator, List[Operator], Array, csr_matrix, None],
        convert_type: bool = True,
    ) -> Array:
        r"""Take an operator out of the frame basis, i.e. return
        ``self.frame_basis @ to_array(op) @ self.frame_basis_adjoint``.

        Args:
            op: The operator or array of operators.
            convert_type: Whether or not to initially convert ``op`` into an expected type.
                          Should only be set to ``False`` in situations in which it is gauranteed
                          that ``op`` is a handled input type.
        Returns:
            Array: The operator in the frame basis.
        """
        if convert_type:
            op = to_numeric_matrix_type(op)

        if self.frame_basis is None or op is None:
            return op

        if isinstance(op, list):
            return [self.operator_out_of_frame_basis(x, convert_type=False) for x in op]
        elif type(op).__name__ == "BCOO":
            return self.frame_basis @ jsparse_matmul(op, self.frame_basis_adjoint.data)
        else:
            # parentheses are necessary for sparse op evaluation
            return self.frame_basis @ (op @ self.frame_basis_adjoint)

    def state_into_frame(
        self,
        t: float,
        y: Array,
        y_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
    ):
        """Take a state into the rotating frame, i.e. return exp(-tF) @ y.

        Args:
            t: Time.
            y: State (array of appropriate size).
            y_in_frame_basis: Whether or not the array y is already in
                              the basis in which the frame is diagonal.
            return_in_frame_basis: Whether or not to return the result
                                   in the frame basis.

        Returns:
            Array: State in frame.
        """
        y = to_numeric_matrix_type(y)
        if self._frame_operator is None:
            return y

        out = y

        # if not in frame basis convert it
        if not y_in_frame_basis:
            out = self.state_into_frame_basis(out)

        # go into the frame using fast diagonal matrix multiplication
        out = (np.exp(self.frame_diag * (-t)) * out.transpose()).transpose()  # = e^{tF}out

        # if output is requested to not be in the frame basis, convert it
        if not return_in_frame_basis:
            out = self.state_out_of_frame_basis(out)

        return out

    def state_out_of_frame(
        self,
        t: float,
        y: Array,
        y_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
    ) -> Array:
        r"""Take a state out of the rotating frame, i.e. ``return exp(tF) @ y``.

        Calls ``self.state_into_frame`` with time reversed.

        Args:
            t: Time.
            y: State (array of appropriate size).
            y_in_frame_basis: Whether or not the array y is already in
                              the basis in which the frame is diagonal.
            return_in_frame_basis: Whether or not to return the result
                                   in the frame basis.

        Returns:
            Array: State out of frame.
        """
        return self.state_into_frame(-t, y, y_in_frame_basis, return_in_frame_basis)

    def _conjugate_and_add(
        self,
        t: float,
        operator: Union[Array, csr_matrix],
        op_to_add_in_fb: Optional[Array] = None,
        operator_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
        vectorized_operators: Optional[bool] = False,
    ) -> Union[Array, csr_matrix]:
        r"""General helper function for computing :math:`\exp(-tF)G\exp(tF) + B`.

        Note: B is added in the frame basis before any potential final change
        out of the frame basis.

        Note: There are two conventions for passing multiple operators at the same
        time. For evaluation with vectorized_operators=False, these operators should
        be passed as (k, dim, dim) Array objects, with the :math:`i^{th}` operator
        being stored as the [i,:,:] entry. For vectorized_operators = True, these
        (vectorized) operators should be passed as a (dim**2, k) Array, with the
        :math:`i^{th}` vectorized operator stored as the [:,i] entry.

        Args:
            t: Time.
            operator: The operator G.
            op_to_add_in_fb: The additional operator B.
            operator_in_fame_basis: Whether ``operator`` is already in the basis
                in which the frame operator is diagonal.
            vectorized_operators: Whether ``operator`` is passed as a vectorized,
                ``(dim^2,)`` Array, rather than a ``(dim,dim)`` Array.

        Returns:
            Array of newly conjugated operator.
        """
        operator = to_numeric_matrix_type(operator)
        op_to_add_in_fb = to_numeric_matrix_type(op_to_add_in_fb)

        if vectorized_operators:
            # If passing vectorized operator, undo vectorization temporarily
            if self._frame_operator is None:
                if op_to_add_in_fb is None:
                    return operator
                else:
                    return operator + op_to_add_in_fb
            if len(operator.shape) == 2:
                operator = operator.T
            operator = operator.reshape(operator.shape[:-1] + (self.dim, self.dim), order="F")

        if self._frame_operator is None:
            if op_to_add_in_fb is None:
                return operator
            else:
                if op_to_add_in_fb is not None:
                    if issparse(operator):
                        op_to_add_in_fb = csr_matrix(op_to_add_in_fb)
                    elif type(operator).__name__ == "BCOO":
                        op_to_add_in_fb = to_BCOO(op_to_add_in_fb)

                return operator + op_to_add_in_fb

        out = operator

        # if not in frame basis convert it
        if not operator_in_frame_basis:
            out = self.operator_into_frame_basis(out)

        # get frame transformation matrix in diagonal basis
        # assumption that F is anti-Hermitian implies conjugation of
        # diagonal gives inversion
        exp_freq = np.exp(self.frame_diag * t)
        frame_mat = exp_freq.conj().reshape(self.dim, 1) * exp_freq
        if issparse(out):
            out = out.multiply(frame_mat)
        elif type(out).__name__ == "BCOO":
            out = out * frame_mat.data
        else:
            out = frame_mat * out

        if op_to_add_in_fb is not None:
            if issparse(out):
                op_to_add_in_fb = csr_matrix(op_to_add_in_fb)
            elif type(out).__name__ == "BCOO":
                op_to_add_in_fb = to_BCOO(op_to_add_in_fb)

            out = out + op_to_add_in_fb

        # if output is requested to not be in the frame basis, convert it
        if not return_in_frame_basis:
            out = self.operator_out_of_frame_basis(out)

        if vectorized_operators:
            # If a vectorized output is required, reshape correctly
            out = out.reshape(out.shape[:-2] + (self.dim**2,), order="F")
            if len(out.shape) == 2:
                out = out.T

        return out

    def operator_into_frame(
        self,
        t: float,
        operator: Union[Operator, Array, csr_matrix],
        operator_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
        vectorized_operators: Optional[bool] = False,
    ) -> Array:
        r"""Bring an operator into the frame, i.e. return
        ``exp(-tF) @ operator @ exp(tF)``.

        Default implementation is to use ``self._conjugate_and_add``.

        Args:
            t: Time.
            operator: Array of appropriate size.
            operator_in_frame_basis: Whether or not the operator is already in
                              the basis in which the frame is diagonal.
            return_in_frame_basis: Whether or not to return the result
                                   in the frame basis.
            vectorized_operators: Whether ``operator`` is passed as a vectorized,
                ``(dim^2,)`` Array, rather than a ``(dim,dim)`` Array.

        Returns:
            Array: operator in frame.
        """
        return self._conjugate_and_add(
            t,
            operator,
            operator_in_frame_basis=operator_in_frame_basis,
            return_in_frame_basis=return_in_frame_basis,
            vectorized_operators=vectorized_operators,
        )

    def operator_out_of_frame(
        self,
        t: float,
        operator: Union[Operator, Array, csr_matrix],
        operator_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
        vectorized_operators: Optional[bool] = False,
    ):
        r"""Bring an operator into the rotating frame, i.e. return
        ``exp(tF) @ operator @ exp(-tF)``.

        Default implmentation is to use `self.operator_into_frame`.

        Args:
            t: Time.
            operator: Array of appropriate size.
            operator_in_frame_basis: Whether or not the operator is already in
                              the basis in which the frame is diagonal.
            return_in_frame_basis: Whether or not to return the result
                                   in the frame basis.
            vectorized_operators: Whether ``operator`` is passed as a vectorized,
                ``(dim^2,)`` Array, rather than a ``(dim,dim)`` Array.

        Returns:
            Array: operator out of frame.
        """
        return self.operator_into_frame(
            -t,
            operator,
            operator_in_frame_basis=operator_in_frame_basis,
            return_in_frame_basis=return_in_frame_basis,
            vectorized_operators=vectorized_operators,
        )

    def generator_into_frame(
        self,
        t: float,
        operator: Union[Operator, Array, csr_matrix],
        operator_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
        vectorized_operators: Optional[bool] = False,
    ):
        r"""Take an generator into the rotating frame, i.e. return
        ``exp(-tF) @ operator @ exp(tF) - F``.

        Default implementation is to use `self._conjugate_and_add`.

        Args:
            t: Time.
            operator: Generator (array of appropriate size).
            operator_in_frame_basis: Whether or not the generator is already in
                              the basis in which the frame is diagonal.
            return_in_frame_basis: Whether or not to return the result
                                   in the frame basis.
            vectorized_operators: Whether ``operator`` is passed as a vectorized,
                ``(dim^2,)`` Array, rather than a ``(dim,dim)`` Array.

        Returns:
            Array: Generator in frame.
        """
        if self.frame_operator is None:
            return to_numeric_matrix_type(operator)
        else:
            # conjugate and subtract the frame diagonal
            return self._conjugate_and_add(
                t,
                operator,
                op_to_add_in_fb=-np.diag(self.frame_diag),
                operator_in_frame_basis=operator_in_frame_basis,
                return_in_frame_basis=return_in_frame_basis,
                vectorized_operators=vectorized_operators,
            )

    def generator_out_of_frame(
        self,
        t: float,
        operator: Union[Operator, Array, csr_matrix],
        operator_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
    ) -> Array:
        r"""Take an operator out of the frame, i.e. return
        ``exp(tF) @ operator @ exp(-tF) + F``.

        Default implementation is to use `self._conjugate_and_add`.

        Args:
            t: Time
            operator: Generator (array of appropriate size).
            operator_in_frame_basis: Whether or not the operator is already in
                              the basis in which the frame is diagonal.
            return_in_frame_basis: Whether or not to return the result
                                   in the frame basis.

        Returns:
            Array: Generator out of frame.
        """
        if self.frame_operator is None:
            return to_numeric_matrix_type(operator)
        else:
            # conjugate and add the frame diagonal
            return self._conjugate_and_add(
                -t,
                operator,
                op_to_add_in_fb=Array(np.diag(self.frame_diag)),
                operator_in_frame_basis=operator_in_frame_basis,
                return_in_frame_basis=return_in_frame_basis,
            )

    @property
    def vectorized_frame_basis(self):
        """Lazily evaluated operator for mapping vectorized operators into the frame basis."""

        if self.frame_basis is None:
            return None

        if self._vectorized_frame_basis is None:
            self._vectorized_frame_basis = np.kron(self.frame_basis.conj(), self.frame_basis)
            self._vectorized_frame_basis_adjoint = self._vectorized_frame_basis.conj().transpose()

        return self._vectorized_frame_basis

    @property
    def vectorized_frame_basis_adjoint(self):
        """Lazily evaluated operator for mapping vectorized operators out of the frame basis."""

        if self.frame_basis is None:
            return None

        if self._vectorized_frame_basis_adjoint is None:
            # trigger lazy evaluation of vectorized_frame_basis
            # pylint: disable=pointless-statement
            self.vectorized_frame_basis

        return self._vectorized_frame_basis_adjoint

    def vectorized_map_into_frame(
        self,
        time: float,
        op: Array,
        operator_in_frame_basis: Optional[bool] = False,
        return_in_frame_basis: Optional[bool] = False,
    ) -> Array:
        r"""Given an operator `op` of dimension `dim**2` assumed to represent vectorized linear map
        in column stacking convention, returns:

        .. math::
            ((e^{tF})^T \otimes e^{-tF}) \times op \times ((e^{-tF})^T \otimes e^{tF}).

        Utilizes element-wise multiplication :math:`op \to \Delta\otimes\bar{\Delta} \odot op`,
        where :math:`\Delta_{ij}=\exp((-d_i+d_j)t)` is the frame difference matrix, as well as
        caches array :math:`\bar{C}\otimes C`, where ``C = self.frame_basis`` for future use.

        Args:
            time: The time t.
            op: The (dim^2,dim^2) Array.
            operator_in_frame_basis: Whether the operator is in the frame basis.
            return_in_frame_basis: Whether the operator should be returned in the
                frame basis.
        Returns:
            op in the frame.
        """
        if self.frame_diag is not None:
            # Put the vectorized operator into the frame basis
            if not operator_in_frame_basis and self.frame_basis is not None:
                op = self.vectorized_frame_basis_adjoint @ (op @ self.vectorized_frame_basis)

            expvals = np.exp(self.frame_diag * time)  # = e^{td_i} = e^{it*Im(d_i)}
            # = kron(e^{-it*Im(d_i)},e^{it*Im(d_i)}), but ~3x faster
            temp_outer = (expvals.conj().reshape(self.dim, 1) * expvals).flatten()
            delta_bar_otimes_delta = np.outer(
                temp_outer.conj(), temp_outer
            )  # = kron(delta.conj(),delta) but >3x faster
            if issparse(op):
                op = op.multiply(delta_bar_otimes_delta)
            else:
                op = delta_bar_otimes_delta * op  # hadamard product

            if not return_in_frame_basis and self.frame_basis is not None:
                op = self.vectorized_frame_basis @ (op @ self.vectorized_frame_basis_adjoint)

        return op


def _is_herm_or_anti_herm(mat: Array, atol: Optional[float] = 1e-10, rtol: Optional[float] = 1e-10):
    r"""Given `mat`, the logic of this function is:
        - if `mat` is hermitian, return `-1j * mat`
        - if `mat` is anti-hermitian, return `mat`
        - otherwise:
            - if `mat.backend == 'jax'` return `jnp.inf * mat`
            - otherwise raise an error

    The main purpose of this function is to hide the pecularities of the
    implementing the above logic in a compileable way in `jax`.

    Args:
        mat: array to check
        atol: absolute tolerance
        rtol: relative tolerance

    Returns:
        Array: anti-hermitian version of `mat` if applicable

    Raises:
        ImportError: if backend is jax and jax is not installed.
        QiskitError: if `mat` is not Hermitian or anti-Hermitian
    """
    mat = to_array(mat)
    mat = Array(mat, dtype=complex)

    if mat.backend == "jax":

        from jax.lax import cond

        mat = mat.data

        if mat.ndim == 1:
            # this function checks if pure imaginary. If yes it returns the
            # array, otherwise it multiplies it by jnp.nan to raise an error
            # Note: pathways in conditionals in jax cannot raise Exceptions
            def anti_herm_conditional(b):
                aherm_pred = jnp.allclose(b, -b.conj(), atol=atol, rtol=rtol)
                return cond(aherm_pred, lambda A: A, lambda A: jnp.nan * A, b)

            # Check if it is purely real, if not apply anti_herm_conditional
            herm_pred = jnp.allclose(mat, mat.conj(), atol=atol, rtol=rtol)
            return Array(cond(herm_pred, lambda A: -1j * A, anti_herm_conditional, mat))
        else:
            # this function checks if anti-hermitian, if yes returns the array,
            # otherwise it multiplies it by jnp.nan
            def anti_herm_conditional(b):
                aherm_pred = jnp.allclose(b, -b.conj().transpose(), atol=atol, rtol=rtol)
                return cond(aherm_pred, lambda A: A, lambda A: jnp.nan * A, b)

            # the following lines check if a is hermitian, otherwise it feeds
            # it into the anti_herm_conditional
            herm_pred = jnp.allclose(mat, mat.conj().transpose(), atol=atol, rtol=rtol)
            return Array(cond(herm_pred, lambda A: -1j * A, anti_herm_conditional, mat))

    else:
        if mat.ndim == 1:
            if np.allclose(mat, mat.conj(), atol=atol, rtol=rtol):
                return -1j * mat
            elif np.allclose(mat, -mat.conj(), atol=atol, rtol=rtol):
                return mat
        else:
            if is_hermitian_matrix(mat, rtol=rtol, atol=atol):
                return -1j * mat
            elif is_hermitian_matrix(1j * mat, rtol=rtol, atol=atol):
                return mat

        # raise error if execution has made it this far
        raise QiskitError("""frame_operator must be either a Hermitian or anti-Hermitian matrix.""")



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

"""
Hamiltonian models module.
"""

from typing import Union, List, Optional
import numpy as np
from scipy.sparse import issparse
from scipy.sparse.csr import csr_matrix
from scipy.sparse.linalg import norm as spnorm

from qiskit import QiskitError
from qiskit.quantum_info.operators import Operator
from qiskit_dynamics.array import Array
from qiskit_dynamics.signals import Signal, SignalList
from qiskit_dynamics.type_utils import to_numeric_matrix_type, to_array
from .generator_model import (
    GeneratorModel,
    transfer_static_operator_between_frames,
    transfer_operators_between_frames,
    construct_operator_collection,
)
from .rotating_frame import RotatingFrame


class HamiltonianModel(GeneratorModel):
    r"""A model of a Hamiltonian for the Schrodinger equation.

    This class represents a Hamiltonian as a time-dependent decomposition the form:

    .. math::
        H(t) = H_d + \sum_j s_j(t) H_j,

    where :math:`H_j` are Hermitian operators, :math:`H_d` is the static component,
    and the :math:`s_j(t)` are either :class:`~qiskit_dynamics.signals.Signal` objects or
    are numerical constants. Constructing a :class:`~qiskit_dynamics.models.HamiltonianModel`
    requires specifying the above decomposition, e.g.:

    .. code-block:: python

        hamiltonian = HamiltonianModel(static_operator=static_operator,
                                       operators=operators,
                                       signals=signals)

    This class inherits most functionality from :class:`GeneratorModel`,
    with the following modifications:

        * The operators :math:`H_d` and :math:`H_j` are assumed and verified to be Hermitian.
        * Rotating frames are dealt with assuming the structure of the Schrodinger
          equation. I.e. Evaluating the Hamiltonian :math:`H(t)` in a
          frame :math:`F = -iH_0`, evaluates the expression
          :math:`e^{-tF}H(t)e^{tF} - H_0`.
    """

    def __init__(
        self,
        static_operator: Optional[Array] = None,
        operators: Optional[List[Operator]] = None,
        signals: Optional[Union[SignalList, List[Signal]]] = None,
        rotating_frame: Optional[Union[Operator, Array, RotatingFrame]] = None,
        in_frame_basis: bool = False,
        evaluation_mode: str = "dense",
        validate: bool = True,
    ):
        """Initialize, ensuring that the operators are Hermitian.

        Args:
            static_operator: Time-independent term in the Hamiltonian.
            operators: List of Operator objects.
            signals: List of coefficients :math:`s_i(t)`. Not required at instantiation, but
                     necessary for evaluation of the model.
            rotating_frame: Rotating frame operator.
                            If specified with a 1d array, it is interpreted as the
                            diagonal of a diagonal matrix. Assumed to store
                            the antihermitian matrix F = -iH.
            in_frame_basis: Whether to represent the model in the basis in which the rotating
                            frame operator is diagonalized.
            evaluation_mode: Evaluation mode to use. Supported options are ``'dense'`` and
                             ``'sparse'``. Call ``help(HamiltonianModel.evaluation_mode)`` for more
                             details.
            validate: If True check input operators are Hermitian.

        Raises:
            QiskitError: if operators are not Hermitian
        """

        # verify operators are Hermitian, and if so instantiate
        operators = to_numeric_matrix_type(operators)
        static_operator = to_numeric_matrix_type(static_operator)

        if validate:
            if (operators is not None) and (not is_hermitian(operators)):
                raise QiskitError("""HamiltonianModel operators must be Hermitian.""")
            if (static_operator is not None) and (not is_hermitian(static_operator)):
                raise QiskitError("""HamiltonianModel static_operator must be Hermitian.""")

        super().__init__(
            static_operator=static_operator,
            operators=operators,
            signals=signals,
            rotating_frame=rotating_frame,
            in_frame_basis=in_frame_basis,
            evaluation_mode=evaluation_mode,
        )

    @property
    def rotating_frame(self) -> RotatingFrame:
        return super().rotating_frame

    @rotating_frame.setter
    def rotating_frame(self, rotating_frame: Union[Operator, Array, RotatingFrame]) -> Array:
        """Sets frame. RotatingFrame objects will always store antihermitian F = -iH.
        The static_operator needs to be adjusted by -H in the new frame."""
        new_frame = RotatingFrame(rotating_frame)

        # convert static operator to new frame setup
        static_op = self._get_static_operator(in_frame_basis=True)
        if static_op is not None:
            static_op = -1j * static_op

        new_static_operator = transfer_static_operator_between_frames(
            static_op,
            new_frame=new_frame,
            old_frame=self.rotating_frame,
        )

        if new_static_operator is not None:
            new_static_operator = 1j * new_static_operator

        # convert operators to new frame set up
        new_operators = transfer_operators_between_frames(
            self._get_operators(in_frame_basis=True),
            new_frame=new_frame,
            old_frame=self.rotating_frame,
        )

        self._rotating_frame = new_frame

        self._operator_collection = construct_operator_collection(
            self.evaluation_mode, new_static_operator, new_operators
        )

    def evaluate(self, time: float) -> Array:
        return -1j * super().evaluate(time)

    def evaluate_rhs(self, time: float, y: Array) -> Array:
        return -1j * super().evaluate_rhs(time, y)


def is_hermitian(
    operators: Union[Array, csr_matrix, List[csr_matrix], "BCOO"], tol: Optional[float] = 1e-10
) -> bool:
    """Validate that operators are Hermitian.

    Args:
        operators: Either a 2d array representing a single operator, a 3d array
                   representing a list of operators, a csr_matrix, or a list
                   of csr_matrix.
        tol: Tolerance for checking zeros.

    Returns:
        bool: Whether or not the operators are Hermitian to within tolerance.

    Raises:
        QiskitError: If an unexpeted type.
    """
    if isinstance(operators, (np.ndarray, Array)):
        adj = None
        if operators.ndim == 2:
            adj = np.transpose(np.conjugate(operators))
        elif operators.ndim == 3:
            adj = np.transpose(np.conjugate(operators), (0, 2, 1))
        return np.linalg.norm(adj - operators) < tol
    elif issparse(operators):
        return spnorm(operators - operators.conj().transpose()) < tol
    elif isinstance(operators, list) and issparse(operators[0]):
        return all(spnorm(op - op.conj().transpose()) < tol for op in operators)
    elif type(operators).__name__ == "BCOO":
        # fall back on array case for BCOO
        return is_hermitian(to_array(operators))

    raise QiskitError("is_hermitian got an unexpected type.")



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Array Class"""

import copy
from functools import wraps
from types import BuiltinMethodType, MethodType
from typing import Optional, Union, Tuple, Set
from numbers import Number

import numpy
from numpy.lib.mixins import NDArrayOperatorsMixin

from qiskit_dynamics.dispatch.dispatch import Dispatch, asarray

__all__ = ["Array"]


class Array(NDArrayOperatorsMixin):
    """Qiskit Array class.

    This class provides a Numpy compatible wrapper to supported Python
    array libraries. Supported backends are 'numpy' and 'jax'.
    """

    def __init__(
        self,
        data: any,
        dtype: Optional[any] = None,
        order: Optional[str] = None,
        backend: Optional[str] = None,
    ):
        """Initialize an Array container.

        Args:
            data: An array_like input. This can be an object of any type
                  supported by the registered `asarray` method for the
                  specified backend.
            dtype: Optional. The dtype of the returned array. This value
                   must be supported by the specified array backend.
            order: Optional. The array order. This value must be supported
                   by the specified array backend.
            backend: A registered array backend name. If None the
                     default array backend will be used.

        Raises:
            ValueError: if input cannot be converted to an Array.
        """

        # Check if we can override setattr and
        # set _data and _backend directly
        if (
            isinstance(data, numpy.ndarray)
            and _is_numpy_backend(backend)
            and _is_equivalent_numpy_array(data, dtype, order)
        ):
            self.__dict__["_data"] = data
            self.__dict__["_backend"] = "numpy"
            return

        if hasattr(data, "__qiskit_array__"):
            array = data.__qiskit_array__(dtype=dtype, backend=backend)
            if not isinstance(array, Array):
                raise ValueError("object __qiskit_array__ method is not producing an Array")
            self._data = array._data
            self._backend = array._backend
            if dtype or order or (backend and backend != self._backend):
                if backend is None:
                    backend = self._backend
                else:
                    self._backend = backend
                self._data = asarray(self._data, dtype=dtype, order=order, backend=backend)
            return

        # Standard init
        self._data = asarray(data, dtype=dtype, order=order, backend=backend)
        self._backend = backend if backend else Dispatch.backend(self._data, subclass=True)

    @property
    def data(self):
        """Return the wrapped array data object"""
        return self._data

    @data.setter
    def data(self, value):
        """Update the wrapped array data object"""
        self._data[:] = value

    @property
    def backend(self):
        """Return the backend of the wrapped array class"""
        return self._backend

    @backend.setter
    def backend(self, value: str):
        """Set the backend of the wrapped array class"""
        Dispatch.validate_backend(value)
        self._data = asarray(self._data, backend=value)
        self._backend = value

    @classmethod
    def set_default_backend(cls, backend: Union[str, None]):
        """Set the default array backend."""
        if backend is not None:
            Dispatch.validate_backend(backend)
        Dispatch.DEFAULT_BACKEND = backend

    @classmethod
    def default_backend(cls) -> str:
        """Return the default array backend."""
        return Dispatch.DEFAULT_BACKEND

    @classmethod
    def available_backends(cls) -> Set[str]:
        """Return a tuple of available array backends"""
        return Dispatch.REGISTERED_BACKENDS

    def __repr__(self):
        prefix = "Array("
        if self._backend == Dispatch.DEFAULT_BACKEND:
            suffix = ")"
        else:
            suffix = f"backend='{self._backend}')"
        return Dispatch.repr(self.backend)(self._data, prefix=prefix, suffix=suffix)

    def __copy__(self):
        """Return a shallow copy referencing the same wrapped data array"""
        return Array(self._data, backend=self.backend)

    def __deepcopy__(self, memo=None):
        """Return a deep copy with a copy of the wrapped data array"""
        return Array(copy.deepcopy(self._data), backend=self.backend)

    def __iter__(self):
        return iter(self._data)

    def __getstate__(self):
        return {"_data": self._data, "_backend": self._backend}

    def __setstate__(self, state):
        self._backend = state["_backend"]
        self._data = state["_data"]

    def __getitem__(self, key: str) -> any:
        """Return value from wrapped array"""
        return self._data[key]

    def __setitem__(self, key: str, value: any):
        """Return value of wrapped array"""
        self._data[key] = value

    def __setattr__(self, name: str, value: any):
        """Set attribute of wrapped array."""
        if name in ("_data", "data", "_backend", "backend"):
            super().__setattr__(name, value)
        else:
            setattr(self._data, name, value)

    def __getattr__(self, name: str) -> any:
        """Get attribute of wrapped array and convert to an Array."""
        # Call attribute on inner array object
        attr = getattr(self._data, name)

        # If attribute is a function wrap the return values
        if isinstance(attr, (MethodType, BuiltinMethodType)):

            @wraps(attr)
            def wrapped_method(*args, **kwargs):
                return self._wrap(attr(*args, **kwargs))

            return wrapped_method

        # If return object is a backend array wrap result
        return self._wrap(attr)

    def __qiskit_array__(self, dtype=None, backend=None):
        if (backend and backend != self.backend) or (dtype and dtype != self.data.dtype):
            return Array(self.data, dtype=dtype, backend=backend)
        return self

    def __array__(self, dtype=None) -> numpy.ndarray:
        if isinstance(self._data, numpy.ndarray) and (dtype is None or dtype == self._data.dtype):
            return self._data
        return numpy.asarray(self._data, dtype=dtype)

    def __len__(self) -> int:
        return len(self._data)

    def __str__(self) -> str:
        return str(self._data)

    def __bool__(self) -> bool:
        return bool(self._data)

    def __int__(self):
        """Convert size 1 array to an int."""
        if numpy.size(self) != 1:
            raise TypeError("only size-1 Arrays can be converted to Python scalars")
        return int(self._data)

    def __float__(self):
        """Convert size 1 array to a float."""
        if numpy.size(self) != 1:
            raise TypeError("only size-1 Arrays can be converted to Python scalars")
        return float(self._data)

    def __complex__(self):
        """Convert size 1 array to a complex."""
        if numpy.size(self) != 1:
            raise TypeError("only size-1 Arrays can be converted to Python scalars")
        return complex(self._data)

    @staticmethod
    def _wrap(obj: Union[any, Tuple[any]], backend: Optional[str] = None) -> Union[any, Tuple[any]]:
        """Wrap return array backend objects as Array objects"""
        if isinstance(obj, tuple):
            return tuple(
                Array(x, backend=backend) if isinstance(x, Dispatch.REGISTERED_TYPES) else x
                for x in obj
            )
        if isinstance(obj, Dispatch.REGISTERED_TYPES):
            return Array(obj, backend=backend)
        return obj

    @classmethod
    def _unwrap(cls, obj):
        """Unwrap an Array or list of Array objects"""
        if isinstance(obj, Array):
            return obj._data
        if isinstance(obj, tuple):
            return tuple(cls._unwrap(i) for i in obj)
        if isinstance(obj, list):
            return list(cls._unwrap(i) for i in obj)
        return obj

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Dispatcher for numpy ufuncs to support the wrapped array backend."""
        out = kwargs.get("out", tuple())

        for i in inputs + out:
            # Only support operations with instances of REGISTERED_TYPES.
            # Use ArrayLike instead of type(self) for isinstance to
            # allow subclasses that don't override __array_ufunc__ to
            # handle ArrayLike objects.
            if not isinstance(i, Dispatch.REGISTERED_TYPES + (Array, Number)):
                return NotImplemented

        # Defer to the implementation of the ufunc on unwrapped values.
        inputs = self._unwrap(inputs)
        if out:
            kwargs["out"] = self._unwrap(out)

        # Get implementation for backend
        backend = self.backend
        dispatch_func = Dispatch.array_ufunc(backend, ufunc, method)
        if dispatch_func == NotImplemented:
            return NotImplemented
        result = dispatch_func(*inputs, **kwargs)

        # Not sure what this case from Numpy docs is?
        if method == "at":
            return None

        # Wrap array results back into Array objects
        return self._wrap(result, backend=self.backend)

    def __array_function__(self, func, types, args, kwargs):
        """Dispatcher for numpy array function to support the wrapped array backend."""
        if not all(issubclass(t, (Array,) + Dispatch.REGISTERED_TYPES) for t in types):
            return NotImplemented

        # Unwrap function Array arguments
        args = self._unwrap(args)
        out = kwargs.get("out", tuple())
        if out:
            kwargs["out"] = self._unwrap(out)

        # Get implementation for backend
        backend = self.backend
        dispatch_func = Dispatch.array_function(backend, func)
        if dispatch_func == NotImplemented:
            return NotImplemented
        result = dispatch_func(*args, **kwargs)
        return self._wrap(result, backend=self.backend)


def _is_numpy_backend(backend: Optional[str] = None):
    return backend == "numpy" or (not backend and Dispatch.DEFAULT_BACKEND == "numpy")


def _is_equivalent_numpy_array(data: any, dtype: Optional[any] = None, order: Optional[str] = None):
    return (not dtype or dtype == data.dtype) and (
        not order
        or (order == "C" and data.flags["C_CONTIGUOUS"])
        or (order == "F" and data.flags["F_CONTIGUOUS"])
    )
